module convection

  use vtype_module

  implicit none

  public :: ConvCalcRadTendency
  public :: ConvCalcRadConvTendency

contains

  !----------------------------------------------------------------------------

  subroutine ConvCalcRadTendency(                                 &
    & Grav, SurfHeatCap, DelTime,                                 & ! (in)
    & kmax, r_Press, r_Temp,                                      & ! (in)
    & NDelFlxMax,                                                 & ! (in)
    & r_RadUwFlux, r_RadDwFlux, ra_DelRadUwFlux, ra_DelRadDwFlux, & ! (in)
    & r_DelTemp, DelSurfTemp                                      & ! (out)
    & )

#if defined(THERMOPROP_IGCpLSF)
    use thermo_prop_ig_cplsf, only : &
#elif defined(THERMOPROP_vdWEOS)
    use thermo_prop_vdWEOS, only : &
#elif defined(THERMOPROP_GS2016)
    use thermo_prop_mix_gs2016_co2n2, only : &
#elif defined(THERMOPROP_IGCpLMD)
    use thermo_prop_ig_cplmd, only : &
#else
    use thermo_prop_co2_sw1996, only : &
#endif
      & ThermoPropDens,           &
      & ThermoPropCp

    real(dp), intent(in ) :: Grav
    real(dp), intent(in ) :: SurfHeatCap
    real(dp), intent(in ) :: DelTime
    integer , intent(in ) :: kmax
    real(dp), intent(in ) :: r_Press        (0:kmax)
    real(dp), intent(in ) :: r_Temp         (0:kmax)
    integer , intent(in ) :: NDelFlxMax
    real(dp), intent(in ) :: r_RadUwFlux    (0:kmax)
    real(dp), intent(in ) :: r_RadDwFlux    (0:kmax)
    real(DP), intent(in ) :: ra_DelRadUwFlux(0:kmax,-1:NDelFlxMax)
    real(DP), intent(in ) :: ra_DelRadDwFlux(0:kmax,-1:NDelFlxMax)
    real(dp), intent(out) :: r_DelTemp      (0:kmax)
    real(dp), intent(out) :: DelSurfTemp


    ! Local variables
    !
    real(dp) :: Temp
    real(dp) :: Dens
    real(dp) :: z_DelAtmMass (1:kmax)
    real(dp) :: z_Cp         (1:kmax)
    real(dp) :: r_RadFlux    (0:kmax)
    real(DP) :: ra_DelRadFlux(0:kmax,-1:NDelFlxMax)

    real(dp) :: rr_Mat(-1:kmax,-1:kmax)
    real(dp) :: r_Vec (-1:kmax)
    integer  :: k
    integer  :: kr
    integer  :: kc


    do k = 1, kmax
      z_DelAtmMass(k) = ( r_Press(k-1) - r_Press(k) ) / Grav
    end do

    r_RadFlux     = r_RadUwFlux     - r_RadDwFlux
    ra_DelRadFlux = ra_DelRadUwFlux - ra_DelRadDwFlux

    do k = 1, kmax
      Temp    = r_Temp(k)
      Dens    = ThermoPropDens( r_Press(k), r_Temp(k) )
      z_Cp(k) = ThermoPropCp( Dens, Temp )
    end do

    rr_Mat = 0.0d0
    r_Vec  = 0.0d0
    !
    ! Ts
    kr = -1
    !
    kc = -1
    rr_Mat(kr,kc) = SurfHeatCap / DelTime + ra_DelRadFlux(0,kc)
    do kc = 0, NDelFlxMax
      rr_Mat(kr,kc) = + ra_DelRadFlux(0,kc)
    end do
    !
    r_Vec(kr)     = - r_RadFlux(0)
    !
    ! T (k=0)
    !   T0 = ( T2 - T1 ) / ( 2 - 1 ) * ( 0 - 1 ) + T1
    !      = - ( T2 - T1 ) + T1
    !      = 2 * T1 - T2
    !   T0 - 2 * T1 + T2 = 0
    !   r_Temp(0) + r_DelTemp(0) - 2 * ( r_Temp(1) + r_DelTemp(1) )
    !     + r_Temp(2) + r_DelTemp(2) = 0
    !   r_DelTemp(0) - 2 * r_DelTemp(1) + r_DelTemp(2) 
    !     = - ( r_Temp(0) - 2 * r_Temp(1) + r_Temp(2) )
    kr = 0
    kc =   0
    rr_Mat(kr,kc) =   1.0d0
    kc =   1
    rr_Mat(kr,kc) = - 2.0d0
    kc =   2
    rr_Mat(kr,kc) =   1.0d0
    !
    r_Vec(kr)     = - ( r_Temp(0) - 2.0d0 * r_Temp(1) + r_Temp(2) )
    !
    ! T
    do kr = 1, kmax
      kc = -1
      rr_Mat(kr,kc) = ra_DelRadFlux(kr,kc) - ra_DelRadFlux(kr-1,kc)
      do kc = 0, NDelFlxMax
        rr_Mat(kr,kc) = ( ra_DelRadFlux(kr,kc) - ra_DelRadFlux(kr-1,kc) )
      end do
      kc = kr
      rr_Mat(kr,kc) = rr_Mat(kr,kc) + z_DelAtmMass(kr) * z_Cp(kr) / DelTime
      !
      r_Vec(kr) = - ( r_RadFlux(kr) - r_RadFlux(kr-1) )
    end do

    call ludecomp_prep_simple ( kmax+2, rr_Mat )
    call ludecomp_solve_simple( kmax+2, rr_Mat, r_Vec )

    ! r_Vec(-1) = surface temperature
    ! r_Vec(0 ) = enthalpy just above the surface
    ! r_Vec(1 ) = enthalpy at 1st atmospheric level
    ! r_Vec(2 ) = enthalpy at 2nd atmospheric level
    ! ...

    DelSurfTemp  = r_Vec(-1)

    r_DelTemp    = r_Vec(0:kmax)


  end subroutine ConvCalcRadTendency

  !----------------------------------------------------------------------------

  subroutine ConvCalcRadConvTendencyCoreAll(                      &
    & z_FlagConvection,                                           & ! (in)
    & Grav, SurfHeatCap, DelTime,                                 & ! (in)
    & kmax, r_Press, r_Temp, SurfTemp,                            & ! (in)
    & z_VertDelTempUppLevel, z_VertDelTempLowLevel,               & ! (in)
    & NDelFlxMax,                                                 & ! (in)
    & r_RadUwFlux, r_RadDwFlux, ra_DelRadUwFlux, ra_DelRadDwFlux, & ! (in)
    & r_DelTemp, DelSurfTemp, SurfSensFlux, r_SensFlux,           & ! (out)
    & NConvLayer,                                                 & ! (out)
    & a_ksConvLayer, a_keConvLayer                                & ! (out)
    & )

#if defined(THERMOPROP_IGCpLSF)
    use thermo_prop_ig_cplsf, only : &
#elif defined(THERMOPROP_vdWEOS)
    use thermo_prop_vdWEOS, only : &
#elif defined(THERMOPROP_GS2016)
    use thermo_prop_mix_gs2016_co2n2, only : &
#elif defined(THERMOPROP_IGCpLMD)
    use thermo_prop_ig_cplmd, only : &
#else
    use thermo_prop_co2_sw1996, only : &
#endif
      & ThermoPropDens,           &
      & ThermoPropCp

    ! Full level index (mid-poing of layers)
    real(dp), intent(in ) :: Grav
    real(dp), intent(in ) :: SurfHeatCap
    real(dp), intent(in ) :: DelTime
    integer , intent(in ) :: kmax
    logical , intent(in ) :: z_FlagConvection(1:kmax)
    real(dp), intent(in ) :: r_Press          (0:kmax)
    real(dp), intent(in ) :: r_Temp           (0:kmax)
    real(dp), intent(in ) :: SurfTemp

    real(DP), intent(in ) :: z_VertDelTempUppLevel      (1:kmax)
    real(DP), intent(in ) :: z_VertDelTempLowLevel      (1:kmax)

    integer , intent(in ) :: NDelFlxMax
    real(dp), intent(in ) :: r_RadUwFlux      (0:kmax)
    real(dp), intent(in ) :: r_RadDwFlux      (0:kmax)
    real(DP), intent(in ) :: ra_DelRadUwFlux  (0:kmax,-1:NDelFlxMax)
    real(DP), intent(in ) :: ra_DelRadDwFlux  (0:kmax,-1:NDelFlxMax)
    real(dp), intent(out) :: r_DelTemp        (0:kmax)
    real(dp), intent(out) :: DelSurfTemp
    real(DP), intent(out) :: SurfSensFlux
    real(dp), intent(out) :: r_SensFlux(0:kmax)
    integer , intent(out) :: NConvLayer
    integer , intent(out) :: a_ksConvLayer(0:kmax)
    integer , intent(out) :: a_keConvLayer(0:kmax)

    ! Function
    !
!!$    real(DP) :: LOCALPressH2OVapSat
!!$    real(DP) :: LOCALDPressH2OVapSatDTemp

    ! Local variables
    !
    real(dp) :: Temp
    real(dp) :: Dens
    real(dp) :: z_DelAtmMass (1:kmax)
    real(dp) :: z_Cp         (1:kmax)
    real(dp) :: r_RadFlux    (0:kmax)
    real(DP) :: ra_DelRadFlux(0:kmax,-1:NDelFlxMax)
    real(dp) :: rr_Mat(-2:kmax,-2:kmax)
    real(dp) :: r_Vec (-2:kmax)

    logical  :: FlagExtpolAbove
    real(dp) :: VDeltaEnthalpy
    real(dp) :: lnPress
    real(dp) :: Press
!!$    real(dp) :: EnthalpyExtpolAbove
!!$    real(dp) :: EnthalpyExtpolBelow

    real(dp) :: TotalDEnthalpyDt
    real(dp) :: FluxTOA
    real(dp) :: FluxSfc
    real(dp) :: Residual

    real(dp) :: r_RadFluxTentative(0:kmax)

    integer  :: kConvS
    integer  :: kConvE

    integer  :: k
    integer  :: kr
    integer  :: kc


    r_RadFlux     = r_RadUwFlux     - r_RadDwFlux
    ra_DelRadFlux = ra_DelRadUwFlux - ra_DelRadDwFlux


    do k = 1, kmax
      z_DelAtmMass(k) = ( r_Press(k-1) - r_Press(k) ) / Grav
    end do

    do k = 1, kmax
      Temp    = r_Temp(k)
      Dens    = ThermoPropDens( r_Press(k), r_Temp(k) )
      z_Cp(k) = ThermoPropCp( Dens, Temp )
    end do

    ! check layer thickness of convection from the surface
    kConvS = 1
    do k = kConvS+1, kmax
      if ( .not. z_FlagConvection(k) ) then
        exit
      end if
    end do
    kConvE = k - 1

    a_ksConvLayer(0) = kConvS
    a_keConvLayer(0) = kConvE

    ! kc = -2  : sensible heat flux
    ! kc = -1  : surface temperature
    ! kc =  0  : atmospheric temperature at the surface (k=0)
    ! kc =  1  : atmospheric temperature at 1st level (k=1)
    ! kc =  2  : atmospheric temperature at 2nd level (k=2)
    ! ...

    rr_Mat = 0.0d0
    r_Vec  = 0.0d0
    !
    ! Energy conservation
    kr = -2
    kc = -2 ! sensible heat flux
    rr_Mat(kr,kc) = - 1.0d0
    kc = -1
    rr_Mat(kr,kc) = + ( ra_DelRadFlux(kConvE,kc) - ra_DelRadFlux(0,kc) )
    !
    do kc = 0, NDelFlxMax
      rr_Mat(kr,kc) = &
        & + ( ra_DelRadFlux(kConvE,kc) - ra_DelRadFlux(0,kc) )
    end do
    !
    do kc = 1, kConvE
      ! memo : Latent heat is neglected to conserve energy
      rr_Mat(kr,kc) = rr_Mat(kr,kc) &
        & + z_DelAtmMass(kc) * z_Cp(kc) / DelTime
    end do
    !
    r_Vec(kr) = - ( r_RadFlux(kConvE) - r_RadFlux(0) )

    !
    ! Ts
#if defined(FIX_SFCTEMP)
     kr = -1
     kc = -1
     rr_Mat(kr,kc) = 1.0d0
     !
     r_Vec(kr) = 0.0d0
#else
    kr = -1
    kc = -2
    rr_Mat(kr,kc) = + 1.0d0
    kc = -1
    rr_Mat(kr,kc) = + SurfHeatCap / DelTime + ra_DelRadFlux(0,kc)
    do kc = 0, NDelFlxMax
      rr_Mat(kr,kc) = rr_Mat(kr,kc) &
        & + ra_DelRadFlux(0,kc)
    end do
    !
    r_Vec(kr) = - r_RadFlux(0)
#endif
    !
    ! T (k=0)
    !   T0 = Ts
    !   r_Temp(0) + r_DelTemp(0) = SurfTemp + DelSurfTemp
    !   - DelSurfTemp + r_DelTemp(0) = SurfTemp - r_Temp(0)
    kr = 0
    kc =  -1
    rr_Mat(kr,kc) = - 1.0d0
    kc =   0
    rr_Mat(kr,kc) = + 1.0d0    !
    r_Vec(kr)     = SurfTemp - r_Temp(0)
    !
    ! T (1 <= k <= kConvE)
    do kr = 1, kConvE
      ! VDeltaTemp : Vertical temperature difference between two layers
      ! VDeltaTemp
      !   =   z_VertDelTempLowLevel(k)
      !     + z_VertDelTempUppLevel(k)
      !--------
      ! r_Temp(k-1) + r_DelTemp(k-1) + VDeltaTemp
      ! = r_Temp(k) + r_DelTemp(k)
      !--------
      ! - r_DelTemp(k-1) + r_DelTemp(k)
      ! = r_Temp(k-1) - r_Temp(k) + VDeltaTemp
      !
      kc = kr-1
      rr_Mat(kr,kc) = - 1.0d0
      kc = kr
      rr_Mat(kr,kc) =   1.0d0
      !
      r_Vec(kr) = &
        &   r_Temp(kr-1) - r_Temp(kr) &
        & + z_VertDelTempLowLevel(kr) + z_VertDelTempUppLevel(kr)
    end do

    ! convection in atmosphere

    NConvLayer = 0

    k = kConvE + 1
    loop_conv_in_atm : do
      if ( z_FlagConvection(k) ) then
        kConvS = k
        find_top_of_conv : do kConvE = kConvS+1, kmax
          if ( .not. z_FlagConvection(kConvE) ) then
            exit find_top_of_conv
          end if
        end do find_top_of_conv
        kConvE = kConvE-1

        NConvLayer = NConvLayer + 1
        a_ksConvLayer(NConvLayer) = kConvS
        a_keConvLayer(NConvLayer) = kConvE

        ! energy conservation
        kr = kConvS
        kc = -1
        rr_Mat(kr,kc) = &
          & + ( ra_DelRadFlux(kConvE,kc) - ra_DelRadFlux(kConvS-1,kc) )
        !
        do kc = 0, NDelFlxMax
          rr_Mat(kr,kc) = &
            & + ( ra_DelRadFlux(kConvE,kc) - ra_DelRadFlux(kConvS-1,kc) )
        end do
        !
        do kc = kConvS, kConvE
          rr_Mat(kr,kc) = rr_Mat(kr,kc) &
            & + z_DelAtmMass(kc) * z_Cp(kc) / DelTime
        end do
        !
        r_Vec(kr) = - ( r_RadFlux(kConvE) - r_RadFlux(kConvS-1) )

        do kr = kConvS+1, kConvE
          ! same as convection layer which originates from the surface
          kc = kr-1
          rr_Mat(kr,kc) = - 1.0d0
          kc = kr
          rr_Mat(kr,kc) =   1.0d0
          !
          r_Vec(kr) = &
            &   r_Temp(kr-1) - r_Temp(kr) &
            & + z_VertDelTempLowLevel(kr) + z_VertDelTempUppLevel(kr)
        end do

        k = kConvE + 1

      else
        ! stable layer
        kr = k

        kc = -1
        rr_Mat(kr,kc) = + ( ra_DelRadFlux(kr,kc) - ra_DelRadFlux(kr-1,kc) )
        !
        do kc = 0, NDelFlxMax
          rr_Mat(kr,kc) = rr_Mat(kr,kc) &
            & + ( ra_DelRadFlux(kr,kc) - ra_DelRadFlux(kr-1,kc) )
        end do
        !
        kc = kr
        rr_Mat(kr,kc) = rr_Mat(kr,kc) &
          & + z_DelAtmMass(kc) * z_Cp(kc) / DelTime
        !
        r_Vec(kr) = - ( r_RadFlux(kr) - r_RadFlux(kr-1) )

        k = k + 1

      end if
      if ( k > kmax ) then
        exit loop_conv_in_atm
      end if
    end do loop_conv_in_atm


    call ludecomp_prep_simple ( (kmax+3), rr_Mat )
    call ludecomp_solve_simple( (kmax+3), rr_Mat, r_Vec )
    !
    r_DelTemp     = r_Vec(0:kmax)
    !
    DelSurfTemp   = r_Vec(-1)
    !
    SurfSensFlux  = r_Vec(-2)

    ! check energy conservation
    !
    TotalDEnthalpyDt = 0.0d0
    do k = 1, kmax
      TotalDEnthalpyDt = TotalDEnthalpyDt &
        & + r_DelTemp(k) * z_DelAtmMass(k) * z_Cp(k)
    end do
    TotalDEnthalpyDt = TotalDEnthalpyDt / DelTime
    !
    FluxTOA = r_RadFlux(kmax)
    k = -1
    FluxTOA = FluxTOA + ra_DelRadFlux(kmax,k) * DelSurfTemp
    do k = 0, NDelFlxMax
      FluxTOA = FluxTOA &
        & + ra_DelRadFlux(kmax,k) * r_DelTemp(k)
    end do
    !
    FluxSfc = r_RadFlux(0)
    k = -1
    FluxSfc = FluxSfc + ra_DelRadFlux(0,k) * DelSurfTemp
    do k = 0, NDelFlxMax
      FluxSfc = FluxSfc &
        & + ra_DelRadFlux(0,k) * r_DelTemp(k)
    end do
    !
    ! TotalDelEnthalpy = - ( FluxTOA - FluxSfc ) + SurfSensFlux
    Residual = TotalDEnthalpyDt + ( FluxTOA - FluxSfc ) - SurfSensFlux
    !
    if ( abs( Residual / FluxSfc ) > 1.0d-10 ) then
      write( 6, * ) 'CHECK: Residual may be large : ', &
        & TotalDEnthalpyDt, FluxTOA, FluxSfc, SurfSensFlux, Residual
      call flush( 6 )
    end if


    ! Reconstruct fluxes
    !
    r_RadFluxTentative = r_RadFlux
    kc = -1
    r_RadFluxTentative = r_RadFluxTentative &
      & + ra_DelRadFlux(:,kc) * DelSurfTemp
    do kc = 0, NDelFlxMax
      r_RadFluxTentative = r_RadFluxTentative &
        & + ra_DelRadFlux(:,kc) * r_DelTemp(kc)
    end do

!!$      r_DelEnthalpy(k) =                                        &
!!$        &   ( r_RadFluxTentative(k-1) - r_RadFluxTentative(k) ) &
!!$        &     / z_DelAtmMass(k) * DelTime                       &
!!$        & + ( r_SensFlux(k-1) - r_SensFlux(k) )                 &
!!$        &     / z_DelAtmMass(k) * DelTime
!!$      r_DelEnthalpy(k) * z_DelAtmMass(k) / DelTime =            &
!!$        &   ( r_RadFluxTentative(k-1) - r_RadFluxTentative(k) ) &
!!$        & + ( r_SensFlux(k-1) - r_SensFlux(k) )
!!$      r_DelEnthalpy(k) * z_DelAtmMass(k) / DelTime              &
!!$        & - ( r_RadFluxTentative(k-1) - r_RadFluxTentative(k) ) &
!!$        & = ( r_SensFlux(k-1) - r_SensFlux(k) )
!!$      ( r_SensFlux(k-1) - r_SensFlux(k) )                       &
!!$        & =   r_DelEnthalpy(k) * z_DelAtmMass(k) / DelTime      &
!!$        &   - ( r_RadFluxTentative(k-1) - r_RadFluxTentative(k) )

      k = kmax
    r_SensFlux(k) = 0.0d0
    do k = kmax, 1, -1
      r_SensFlux(k-1)                                           &
        & = r_SensFlux(k)                                       &
        & + r_DelTemp(k) * z_DelAtmMass(k) * z_Cp(k) / DelTime        &
        & - ( r_RadFluxTentative(k-1) - r_RadFluxTentative(k) )
    end do
    ! check
    k = 0
    Residual = r_SensFlux(k) - SurfSensFlux
    if ( SurfSensFlux /= 0.0d0 ) then
      if ( abs( Residual / SurfSensFlux ) > 1.0d-10 ) then
        write( 6, * ) 'CHECK: Residual of surface sensible heat flux may be large : ', &
          & r_SensFlux(k), SurfSensFlux, Residual, Residual / SurfSensFlux
        call flush( 6 )
      end if
    end if


  end subroutine ConvCalcRadConvTendencyCoreAll

  !----------------------------------------------------------------------------

  subroutine ConvCalcRadConvTendency(                             &
    & FlagMoistConvection,                                        & ! (in)
    & Grav, SurfHeatCap, DelTime,                                 & ! (in)
    & kmax, r_Press, r_Temp, SurfTemp, r_VMRH2OVap, r_MeanMolWt,  & ! (in)
    & NDelFlxMax,                                                 & ! (in)
    & r_RadUwFlux, r_RadDwFlux, ra_DelRadUwFlux, ra_DelRadDwFlux, & ! (in)
    & r_DelTemp, DelSurfTemp, r_DelVMRH2OVap,                     & ! (out)
    & SurfSensFlux, r_SensFlux,                                   & ! (out)
    & FlagStdOut                                                  & ! (in)
    & )

    logical , intent(in ) :: FlagMoistConvection
    real(dp), intent(in ) :: Grav
    real(dp), intent(in ) :: SurfHeatCap
    real(dp), intent(in ) :: DelTime
    integer , intent(in ) :: kmax
    real(dp), intent(in ) :: r_Press        (0:kmax)
    real(dp), intent(in ) :: r_Temp         (0:kmax)
    real(dp), intent(in ) :: SurfTemp
    real(dp), intent(in ) :: r_VMRH2OVap    (0:kmax)
    real(DP), intent(in ) :: r_MeanMolWt    (0:kmax)
    integer , intent(in ) :: NDelFlxMax
    real(dp), intent(in ) :: r_RadUwFlux    (0:kmax)
    real(dp), intent(in ) :: r_RadDwFlux    (0:kmax)
    real(DP), intent(in ) :: ra_DelRadUwFlux(0:kmax,-1:NDelFlxMax)
    real(DP), intent(in ) :: ra_DelRadDwFlux(0:kmax,-1:NDelFlxMax)
    real(dp), intent(out) :: r_DelTemp      (0:kmax)
    real(dp), intent(out) :: DelSurfTemp
    real(DP), intent(out) :: r_DelVMRH2OVap (0:kmax)
    real(DP), intent(out) :: SurfSensFlux
    real(dp), intent(out) :: r_SensFlux(0:kmax)
    logical , intent(in ), optional :: FlagStdOut


    ! Local variables
    !
    real(dp) :: r_TempTentative                (0:kmax)
    real(dp) :: r_DTempDEnthalpy               (0:kmax)

    real(DP) :: r_DTempDlnPAd              (0:kmax)
    real(DP) :: z_VertDelTempUppLevel      (1:kmax)
    real(DP) :: z_VertDelTempLowLevel      (1:kmax)

    real(DP) :: z_RadFluxConv  (1:kmax)

    ! Full level index (mid-poing of layers)
    integer , parameter :: kConvSfcS = 1
    integer             :: kConvSfcE

    real(dp)            :: z_DelAtmMass (1:kmax)

    real(dp)            :: Press

    integer             :: iItr
    integer             :: NItr = 1000
    logical             :: z_FlagConvection    (1:kmax)
    logical             :: z_FlagConvectionPrev(1:kmax)
    logical             :: FlagFinish

    integer             :: NConvLayer
    integer             :: a_ksConvLayer(0:kmax)
    integer             :: a_keConvLayer(0:kmax)

    integer             :: k


    if ( FlagMoistConvection ) then
      write( 6, * ) 'ERROR : FlagMoistConvection of true is not supported, now.'
      stop
    end if


    do k = 1, kmax
      z_DelAtmMass(k) = ( r_Press(k-1) - r_Press(k) ) / Grav
    end do


    ! Find top of convection
    do k = 1, kmax
      z_RadFluxConv(k) =                                  &
        & - (     r_RadUwFlux(k  ) - r_RadDwFlux(k  )     &
        &     - ( r_RadUwFlux(k-1) - r_RadDwFlux(k-1) ) )
    end do
    do kConvSfcE = 1, kmax
      if ( z_RadFluxConv(kConvSfcE) < 0.0d0 ) exit
    end do
    if ( kConvSfcE > kmax ) then
      write( 6, * ) 'Convection seems to reach top of model from the surface'
      stop
    end if


    call ConvCalcDryConvDTempDlnP(  &
      & kmax, r_Press, r_Temp,      & ! (in )
      & r_DTempDlnPAd               & ! (out)
      & )

    do k = 1, kmax
      if ( r_Press(k) == 0.0d0 ) then
        Press = sqrt( r_Press(k-1) * r_Press(k-1) * 0.9d0 )
      else
        Press = sqrt( r_Press(k-1) * r_Press(k)           )
      end if
      !
      ! These values mean value at upper level minus lower level.
      ! Negative sign mean upper level value (temperature) is lower than
      ! lower level value.
      !
      z_VertDelTempUppLevel(k) = &
        & - r_DTempDlnPAd(k  ) * log( Press        / r_Press(k) )
      z_VertDelTempLowLevel(k) = &
        & - r_DTempDlnPAd(k-1) * log( r_Press(k-1) / Press      )
    end do


    ! Initialization
    z_FlagConvection = .false.
    r_DelTemp = 0.0d0

    iteration : do iItr = 1, NItr

      r_TempTentative = r_Temp + r_DelTemp


      z_FlagConvectionPrev = z_FlagConvection

      ! check unstable layers
      z_FlagConvection(kConvSfcS:kConvSfcE) = .true.
      do k = kConvSfcE+1, kmax
!!$        if (   r_EnthalpyTentative(k-1)+z_VertDelEnthalpy(k  ) &
!!$          &  > r_EnthalpyTentative(k  ) ) then
        if (   r_TempTentative(k-1)         &
          &      + z_VertDelTempLowLevel(k) &
          &      + z_VertDelTempUppLevel(k) &
          &  > r_TempTentative(k  ) ) then
          ! Unstable
          z_FlagConvection(k-1) = .true.
          z_FlagConvection(k  ) = .true.
        end if
      end do
      FlagFinish = .true.
      loop_check_finish : do k = 1, kmax
        if ( z_FlagConvectionPrev(k) .neqv. z_FlagConvection(k) ) then
          FlagFinish = .false.
          exit loop_check_finish
        end if
      end do loop_check_finish
      if ( FlagFinish ) then
        exit iteration
      end if


      ! calc tendency
      call ConvCalcRadConvTendencyCoreAll(                            &
        & z_FlagConvection,                                           &
        & Grav, SurfHeatCap, DelTime,                                 &
        & kmax, r_Press, r_Temp, SurfTemp,                            &
        & z_VertDelTempUppLevel, z_VertDelTempLowLevel,               & ! (in)
        & NDelFlxMax,                                                 &
        & r_RadUwFlux, r_RadDwFlux, ra_DelRadUwFlux, ra_DelRadDwFlux, &
        & r_DelTemp, DelSurfTemp, SurfSensFlux, r_SensFlux,           &
        & NConvLayer,                                                 & ! (out)
        & a_ksConvLayer, a_keConvLayer                                & ! (out)
        & )

    end do iteration
    if ( iItr >= NItr ) then
      write( 6, * ) 'ERROR : convection does not finish.'
      stop
    end if
    if ( present( FlagStdOut ) ) then
      if ( FlagStdOut ) then
        write( 6, * ) 'Convection iterations    : ', iItr
        write( 6, * ) 'Convection from surface  : ', a_keConvLayer(0), 'layers'
        write( 6, * ) 'Convection in atmosphere : ', NConvLayer, 'layers'
        do k = 1, NConvLayer
          write( 6, '(a,i3.3,a,i3.3)' ) '   ', &
            & a_ksConvLayer(k), '-', a_keConvLayer(k)
        end do
        call flush( 6 )
      end if
    end if


    r_DelVMRH2OVap = 0.0d0


  end subroutine ConvCalcRadConvTendency

  !----------------------------------------------------------------------------

  subroutine ConvCalcDryConvDTempDlnP( &
    & kmax, r_Press, r_Temp,           & ! (in )
    & r_DTempDlnP                      & ! (out)
    & )

#if defined(THERMOPROP_IGCpLSF)
    use thermo_prop_ig_cplsf, only : &
#elif defined(THERMOPROP_vdWEOS)
    use thermo_prop_vdWEOS, only : &
#elif defined(THERMOPROP_GS2016)
    use thermo_prop_mix_gs2016_co2n2, only : &
#elif defined(THERMOPROP_IGCpLMD)
    use thermo_prop_ig_cplmd, only : &
#else
    use thermo_prop_co2_sw1996, only : &
#endif
      & ThermoPropDens,                 &
      & ThermoPropAdLapRateDTDP

    integer , intent(in ) :: kmax
    real(DP), intent(in ) :: r_Press          (0:kmax)
    real(DP), intent(in ) :: r_Temp           (0:kmax)
    real(DP), intent(out) :: r_DTempDlnP      (0:kmax)


    ! Local variables
    !
    real(dp) :: Press
    real(dp) :: Temp
    real(dp) :: Dens
    integer  :: k


    do k = 0, kmax
      Press = r_Press(k)
      Temp  = r_Temp (k)
      Dens  = ThermoPropDens( Press, Temp )
      r_DTempDlnP(k) = Press * ThermoPropAdLapRateDTDP( Dens, Temp )
    end do


  end subroutine ConvCalcDryConvDTempDlnP

  !----------------------------------------------------------------------------

end module convection
