! J. Gernert and R. Span, 
! EOS-CG: A Helmholtz energy mixture model for humid gases and CCS mixtures, 
! Journal of Chemical Thermodynamics, 93, 274 (2016);
! https://doi.org/10.1016/j.jct.2015.05.015

! MEMO:
!   Deltas appeared in GS2016 are ratio of number densities.
!   So, the delta_{o,i} for each component is calculated by the use of mass
!   densities, as follows.
!     MolDens  = Dens / MeanMolWeight
!     MolDensC = DensC / MolWeight
!     Delta_{o,i} = MolDens / MolDensC
!                 = Dens / MeanMolWeight / ( DensC / MolWeight )
!                 = Dens / DensC * MolWeight / MeanMolWeight

module thermo_prop_mix_gs2016_co2n2

  use vtype_module

  implicit none

  private

  character(stdstr), parameter :: ModuleName = 'thermo_prop_mix_gs2016_co2n2'

  ! flag for including entropy of mixing
  logical          , parameter :: FlagMixingEffect = .true.
!!$  logical          , parameter :: FlagMixingEffect = .false.

  ! information of compnents
  integer          , parameter :: NComp = 3
  character(stdstr), save      :: m_Name             ( NComp )
  logical          , save      :: m_Flag             ( NComp )
  real(dp)         , save      :: m_UnivGasConst     ( NComp )
  real(dp)         , save      :: m_UnivGasConstRatio( NComp )
  real(dp)         , save      :: m_MolWeight        ( NComp )
  real(dp)         , save      :: m_VMR              ( NComp )
  real(dp)         , save      :: m_TempC            ( NComp )
  real(dp)         , save      :: m_DensC            ( NComp )
  real(dp)         , save      :: m_MolDensC         ( NComp )
  real(dp)         , save      :: TempR
  real(dp)         , save      :: DensR
  real(dp)         , save      :: MolDensR
  real(dp)         , save      :: MeanMolWeight

  ! set order and flag of components
  data m_Name / "CO2"  , "H2O"  , "N2 "  /
  data m_Flag / .true. , .false., .true. /

  logical , save :: FlagIdealGas = .false.
  real(dp), save :: ConvThreshold = 1.0d-11

  data FlagIdealGas  /.false./
  data ConvThreshold /1.0d-11/

  real(dp), save :: mm_ResBetaT (NComp,NComp)
  real(dp), save :: mm_ResGammaT(NComp,NComp)
  real(dp), save :: mm_ResBetaV (NComp,NComp)
  real(dp), save :: mm_ResGammaV(NComp,NComp)
  real(dp), save :: mm_ResF     (NComp,NComp)

  ! beta_{T,i,j} (Table 2)
  !   1 = CO2, 2 = H2O, 3 = N2
  data mm_ResBetaT &
    & / &
    &   -1.0d100,  1.030538d0  ,  0.994140013d0, & ! (*,CO2)
    &   2*-1.0d100             ,  0.954149d0   , & ! (*,H2O)
    &   3*-1.0d100                               & ! (*,N2 )
    & /
  ! gamma_{T,i,j} (Table 2)
  !   1 = CO2, 2 = H2O, 3 = N2
  data mm_ResGammaT &
    & / &
    &   -1.0d100,  0.828472d0,  1.107654104d0, & ! (*,CO2)
    &   2*-1.0d100           ,  0.805147d0   , & ! (*,H2O)
    &   3*-1.0d100                             & ! (*,N2 )
    & /
  ! beta_{v,i,j} (Table 2)
  !   1 = CO2, 2 = H2O, 3 = N2
  data mm_ResBetaV &
    & / &
    &   -1.0d100,  1.021392d0,  1.022709642d0, & ! (*,CO2)
    &   2*-1.0d100           ,  1.079628d0   , & ! (*,H2O)
    &   3*-1.0d100                             & ! (*,N2 )
    & /
  ! gamma_{v,i,j} (Table 2)
  !   1 = CO2, 2 = H2O, 3 = N2
  data mm_ResGammaV &
    & / &
    &   -1.0d100,  0.895156d0, 1.047578256d0, & ! (*,CO2)
    &   2*-1.0d100           , 0.733443d0   , & ! (*,N2 )
    &   3*-1.0d100                            & ! (*,N2 )
    & /
  ! F_{i,j} (Table 2)
  !   1 = CO2, 2 = H2O, 3 = N2
  data mm_ResF &
    & / &
    &   -1.0d100,  1.0d0  ,  1.0d0  , & ! (*,CO2)
    &   2*-1.0d100        ,  1.0d0  , & ! (*,H2O)
    &   3*-1.0d100                    & ! (*,N2 )
    & /


  integer, parameter :: NPolMax = 8

  real(dp), save :: amm_BinDepFuncSmallN(NPolMax,NComp,NComp)
  real(dp), save :: amm_BinDepFuncSmallT(NPolMax,NComp,NComp)
  integer , save :: amm_BinDepFuncSmallD(NPolMax,NComp,NComp)
  integer , save :: amm_BinDepFuncSmallL(NPolMax,NComp,NComp)
  real(dp), save :: amm_BinDepFuncEta   (3:6    ,NComp,NComp)
  real(dp), save :: amm_BinDepFuncEps   (3:6    ,NComp,NComp)
  real(dp), save :: amm_BinDepFuncBeta  (3:6    ,NComp,NComp)
  real(dp), save :: amm_BinDepFuncGamma (3:6    ,NComp,NComp)


  ! Table 3
  data amm_BinDepFuncSmallN &
    & / &
    !   (1:8,1,1) CO2,CO2
    &   8*-1.0d100, &
    !   (1:8,2,1) H2O,CO2
    &   3.9440467d-1, -1.7634732d0 ,  1.4620755d-1, 8.7522320d-3, &
    &   2.0349398d0 , -9.0350250d-2, -2.1638854d-1, 3.9612170d-2, &
    !   (1:8,3,1) N2,CO2
    &   2.8661625028399d-1, -1.0919833861247d-1, -1.1374032082270,  7.6580544237358d-1, &
    &   4.2638000926819d-3,  1.7673538204534d-1, 2*-1.0d100     , &
    !   (1:8,1,2) CO2,H2O
    &   8*-1.0d100, &
    !   (1:8,2,2) H2O,H2O
    &   8*-1.0d100, &
    !   (1:8,3,2) N2 ,H2O
    &    4.0142079d0, -1.1573939d0, -7.2102425d0 , -5.3251223d0, &
    &   -2.2155867d0, 3*-1.0d100  , &
    !   (1:8,1,3) CO2,N2
    &   8*-1.0d100, &
    !   (1:8,2,3) H2O,N2
    &   8*-1.0d100, &
    !   (1:8,3,3) N2 ,N2
    &   8*-1.0d100  &
    & /
  ! Table 3
  data amm_BinDepFuncSmallT &
    & / &
    !   (1:8,1,1) CO2, CO2
    &   8*-1.0d100, &
    !   (1:8,2,1) H2O, CO2
    &   0.880d0, 2.932d0, 2.433d0, 1.330d0, 4.416d0, 5.514d0, 5.203d0, 1.000d0, &
    !   (1:8,3,1) N2 , CO2
    &   1.850d0, 1.400d0, 3.200d0, 2.500d0, 8.000d0, 3.750d0, 2*-1.0d100, &
    !   (1:8,1,2) CO2, H2O
    &   8*-1.0d100, &
    !   (1:8,2,2) H2O, H2O
    &   8*-1.0d100, &
    !   (1:8,3,2) N2 , H2O
    &   0.547d0, 0.055d0, 1.925d0, 0.552d0, 1.000d0, 3*-1.0d100, &
    !   (1:8,1,3) CO2, N2
    &   8*-1.0d100, &
    !   (1:8,2,3) H2O, N2
    &   8*-1.0d100, &
    !   (1:8,3,3) N2 , N2
    &   8*-1.0d100  &
    & /
  ! Table 3
  data amm_BinDepFuncSmallD &
    & / &
    !   (1:8,1,1) CO2, CO2
    &   8*-100, &
    !   (1:8,2,1) H2O, CO2
    &   1, 1, 3, 0, 2, 3, 1, 5, &
    !   (1:8,3,1) N2 , CO2
    &   2, 3, 1, 1, 1, 2, 2*-100, &
    !   (1:8,1,2) CO2, H2O
    &   8*-100, &
    !   (1:8,2,2) H2O, H2O
    &   8*-100, &
    !   (1:8,3,2) N2 , H2O
    &   1, 1, 1, 2, 4, 3*-100, &
    !   (1:8,1,3) CO2, N2
    &   8*-100, &
    !   (1:8,2,3) H2O, N2
    &   8*-100, &
    !   (1:8,3,3) N2 , N2
    &   8*-100  &
    & /
  ! Table 3
  data amm_BinDepFuncSmallL &
    & / &
    !   (1:8,1,1) CO2, CO2
    &   8*-100, &
    !   (1:8,2,1) H2O, CO2
    &   0, 0, 0, 1, 1, 1, 2, 2, &
    !   (1:8,3,1) N2 , CO2
    &   8*-100, &
    !   (1:8,1,2) CO2, H2O
    &   8*-100, &
    !   (1:8,2,2) H2O, H2O
    &   8*-100, &
    !   (1:8,3,2) N2 , H2O
    &   -100, 1, 1, 1, 1, 3*-100, &
    !   (1:8,1,3) CO2, N2
    &   8*-100, &
    !   (1:8,2,3) H2O, N2
    &   8*-100, &
    !   (1:8,3,3) N2 , N2
    &   8*-100  &
    & /
  ! Table 3
  data amm_BinDepFuncEta &
    & / &
    !   (3:6,1,1) CO2, CO2
    &   4*-1.0d100, &
    !   (3:6,2,1) H2O, CO2
    &   4*-1.0d100, &
    !   (3:6,3,1) N2 , CO2
    &   0.25d0, 0.25d0, 0.00d0, 0.00d0, &
    !   (3:6,1,2) CO2, H2O
    &   4*-1.0d100, &
    !   (3:6,2,2) H2O, H2O
    &   4*-1.0d100, &
    !   (3:6,3,2) N2 , H2O
    &   4*-1.0d100, &
    !   (3:6,1,3) CO2, N2
    &   4*-1.0d100, &
    !   (3:6,2,3) H2O, N2
    &   4*-1.0d100, &
    !   (3:6,3,3) N2 , N2
    &   4*-1.0d100  &
    & /
  ! Table 3
  data amm_BinDepFuncEps &
    & / &
    !   (3:6,1,1) CO2, CO2
    &   4*-1.0d100, &
    !   (3:6,2,1) H2O, CO2
    &   4*-1.0d100, &
    !   (3:6,3,1) N2 , CO2
    &   0.5d0, 0.5d0, 0.5d0, 0.5d0, &
    !   (3:6,1,2) CO2, H2O
    &   4*-1.0d100, &
    !   (3:6,2,2) H2O, H2O
    &   4*-1.0d100, &
    !   (3:6,3,2) N2 , H2O
    &   4*-1.0d100, &
    !   (3:6,1,3) CO2, N2
    &   4*-1.0d100, &
    !   (3:6,2,3) H2O, N2
    &   4*-1.0d100, &
    !   (3:6,3,3) N2 , N2
    &   4*-1.0d100  &
    & /
  ! Table 3
  data amm_BinDepFuncBeta &
    & / &
    !   (3:6,1,1) CO2, CO2
    &   4*-1.0d100, &
    !   (3:6,2,1) H2O, CO2
    &   4*-1.0d100, &
    !   (3:6,3,1) N2 , CO2
    &   0.75d0, 1.00d0, 2.00d0, 3.00d0, &
    !   (3:6,1,2) CO2, H2O
    &   4*-1.0d100, &
    !   (3:6,2,2) H2O, H2O
    &   4*-1.0d100, &
    !   (3:6,3,2) N2 , H2O
    &   4*-1.0d100, &
    !   (3:6,1,3) CO2, N2
    &   4*-1.0d100, &
    !   (3:6,2,3) H2O, N2
    &   4*-1.0d100, &
    !   (3:6,3,3) N2 , N2
    &   4*-1.0d100  &
    & /
  ! Table 3
  data amm_BinDepFuncGamma &
    & / &
    !   (3:6,1,1) CO2, CO2
    &   4*-1.0d100, &
    !   (3:6,2,1) H2O, CO2
    &   4*-1.0d100, &
    !   (3:6,3,1) N2 , CO2
    &   0.5d0, 0.5d0, 0.5d0, 0.5d0, &
    !   (3:6,1,2) CO2, H2O
    &   4*-1.0d100, &
    !   (3:6,2,2) H2O, H2O
    &   4*-1.0d100, &
    !   (3:6,3,2) N2 , H2O
    &   4*-1.0d100, &
    !   (3:6,1,3) CO2, N2
    &   4*-1.0d100, &
    !   (3:6,2,3) H2O, N2
    &   4*-1.0d100, &
    !   (3:6,3,3) N2 , N2
    &   4*-1.0d100  &
    & /

  ! Eq. (9)
  ! NOTE : K_{pol,i,j}, K_{exp,i,j}, K_{spec,i,j} are number of terms. 
  ! mm_KPol  = K_{pol,i,j}  : the number of terms including n, d, t
  ! mm_KExp  = K_{exp,i,j}  : the number of terms including n, d, t, l
  ! mm_KSpec = K_{spec,i,j} : the number of terms including n, d, t, eta, eps, beta, gamma
  integer , save :: mm_KPol (NComp,NComp)
  integer , save :: mm_KExp (NComp,NComp)
  integer , save :: mm_KSpec(NComp,NComp)

  data mm_KPol &
    & / &
    !   (1,1) CO2, CO2  /  (2,1) H2O, CO2  /  (3,1) N2 , CO2
    &   1*-100          ,  0               ,  2              , &
    !   (1,2) CO2, H2O  /  (2,2) H2O, H2O  /  (3,2) N2 , H2O
    &   2*-100                             ,  1              , &
    !   (1,3) CO2, N2   /  (2,3) H2O, N2   /  (3,3) N2 , N2
    &   3*-100                                                 &
    & /
  data mm_KExp &
    & / &
    !   (1,1) CO2, CO2  /  (2,1) H2O, CO2  /  (3,1) N2 , CO2
    &   1*-100          ,  8               ,  0              , &
    !   (1,2) CO2, H2O  /  (2,2) H2O, H2O  /  (3,2) N2 , H2O
    &   2*-100                             ,  4              , &
    !   (1,3) CO2, N2   /  (2,3) H2O, N2   /  (3,3) N2 , N2
    &   3*-100                                                 &
    & /
  data mm_KSpec &
    & / &
    !   (1,1) CO2, CO2  /  (2,1) H2O, CO2  /  (3,1) N2 , CO2
    &   1*-100          ,  0               ,  4              , &
    !   (1,2) CO2, H2O  /  (2,2) H2O, H2O  /  (3,2) N2 , H2O
    &   2*-100                             ,  0              , &
    !   (1,3) CO2, N2   /  (2,3) H2O, N2   /  (3,3) N2 , N2
    &   3*-100                                                 &
    & /

  real(dp), parameter :: UnivGasConst = 8.314510d0  ! J mole-1 K-1
  real(dp), save      :: GasConst


  ! UPDATED
  public :: ThermoPropInit
  public :: ThermoPropFinalization
  public :: ThermoPropInqUnivGasConst
  public :: ThermoPropInqMolWeight
  public :: ThermoPropHelmholtzEne
  public :: ThermoPropCompFact
  public :: ThermoPropDens
  public :: ThermoPropTemp
  public :: ThermoPropTempFromPressEnthalpy
  public :: ThermoPropEntropy
  public :: ThermoPropEnthalpy
  public :: ThermoPropInternalEnergy
  public :: ThermoPropCp
  public :: ThermoPropCv
  public :: ThermoPropDPressDDensConstTemp
  public :: ThermoPropDPressDTempConstDens
  public :: ThermoPropThermoExpCoef
  public :: ThermoPropAdLapRateDTDP
  public :: ThermoPropDensTempFromPressEnthalpy
  public :: ThermoPropJudgePhase
  public :: ThermoPropJudgePhaseGC


  !----------------------------------------------------------------------------

contains

  !----------------------------------------------------------------------------
  ! set parameters
  !----------------------------------------------------------------------------

  subroutine ThermoPropInit( &
    & NSubst, m_SubstName, m_SubstVMR, &
    & FlagIdealGasSet, ConvThresholdSet  &
    & )

    use thermo_prop_co2_sw1996, only :                              &
      & CO2_ThermoPropInit            => ThermoPropInit,            &
      & CO2_ThermoPropInqUnivGasConst => ThermoPropInqUnivGasConst, &
      & CO2_ThermoPropInqMolWeight    => ThermoPropInqMolWeight,    &
      & CO2_ThermoPropInqTempCrit     => ThermoPropInqTempCrit,     &
      & CO2_ThermoPropInqDensCrit     => ThermoPropInqDensCrit
    use thermo_prop_n2_s2000, only :                               &
      & N2_ThermoPropInit            => ThermoPropInit,            &
      & N2_ThermoPropInqUnivGasConst => ThermoPropInqUnivGasConst, &
      & N2_ThermoPropInqMolWeight    => ThermoPropInqMolWeight,    &
      & N2_ThermoPropInqTempCrit     => ThermoPropInqTempCrit,     &
      & N2_ThermoPropInqDensCrit     => ThermoPropInqDensCrit


    integer     , intent(in)           :: NSubst
    character(*), intent(in)           :: m_SubstName(NSubst)
    real(dp)    , intent(in)           :: m_SubstVMR (NSubst)
    logical     , intent(in), optional :: FlagIdealGasSet
    real(dp)    , intent(in), optional :: ConvThresholdSet


    ! Local variables
    !
    real(dp) :: SumVMR
    integer  :: m
    integer  :: n


    ! copy values in tables to transpose positions
    do m = 1, NComp
      do n = m+1, NComp
        mm_ResBetaT           (m,n) = mm_ResBetaT           (n,m)
        mm_ResGammaT          (m,n) = mm_ResGammaT          (n,m)
        mm_ResBetaV           (m,n) = mm_ResBetaV           (n,m)
        mm_ResGammaV          (m,n) = mm_ResGammaV          (n,m)
        mm_ResF               (m,n) = mm_ResF               (n,m)
        amm_BinDepFuncSmallN(:,m,n) = amm_BinDepFuncSmallN(:,n,m)
        amm_BinDepFuncSmallT(:,m,n) = amm_BinDepFuncSmallT(:,n,m)
        amm_BinDepFuncSmallD(:,m,n) = amm_BinDepFuncSmallD(:,n,m)
        amm_BinDepFuncSmallL(:,m,n) = amm_BinDepFuncSmallL(:,n,m)
        amm_BinDepFuncEta   (:,m,n) = amm_BinDepFuncEta   (:,n,m)
        amm_BinDepFuncEps   (:,m,n) = amm_BinDepFuncEps   (:,n,m)
        amm_BinDepFuncBeta  (:,m,n) = amm_BinDepFuncBeta  (:,n,m)
        amm_BinDepFuncGamma (:,m,n) = amm_BinDepFuncGamma (:,n,m)
        mm_KPol               (m,n) = mm_KPol               (n,m)
        mm_KExp               (m,n) = mm_KExp               (n,m)
        mm_KSpec              (m,n) = mm_KSpec              (n,m)
      end do
    end do


    ! check
    if ( NSubst /= NComp ) then
      write( 6, * ) 'ERROR: In ', trim( ModuleName )
      write( 6, * ) 'ERROR: The current program supports the case of NSubst = 3.'
      stop
    end if
    do m = 1, NComp
      if ( trim( m_SubstName(m) ) /= m_Name(m) ) then
        write( 6, * ) 'ERROR: In ', trim( ModuleName )
        write( 6, * ) 'ERROR: The current program supports only the case below : '
        write( 6, * ) '         m_SubstName(', m, ') = "' // trim(m_Name(m)) // '"'
        stop
      end if
    end do

    do m = 1, NComp
      if ( m_Flag(m) ) then
        m_VMR(m) = m_SubstVMR(m)
      else
        m_VMR(m) = -1.0d100
      end if
    end do
    SumVMR = 0.0d0
    do m = 1, NComp
      if ( m_Flag(m) ) then
        SumVMR = SumVMR + m_VMR(m)
      end if
    end do
    if ( abs( 1.0d0 - SumVMR ) > 1.0d-10 ) then
      write( 6, * ) 'ERROR: In ', trim( ModuleName )
      write( 6, * ) 'ERROR: Summation of VMR does not match to 1.0, ', SumVMR
      stop
    end if


    if ( present( FlagIdealGasSet ) ) then
      FlagIdealGas = FlagIdealGasSet
    end if

    if ( present( ConvThresholdSet ) ) then
      ConvThreshold = ConvThresholdSet
    end if

    ! initialization of module used in this module
    call CO2_ThermoPropInit( &
      & FlagIdealGasSet, ConvThresholdSet  &
      & )
    call N2_ThermoPropInit( &
      & FlagIdealGasSet, ConvThresholdSet  &
      & )


    ! parameters of each component
    m = 0
    ! CO2
    m = m + 1
    m_UnivGasConst(m) = CO2_ThermoPropInqUnivGasConst()
    m_MolWeight   (m) = CO2_ThermoPropInqMolWeight()
    m_TempC       (m) = CO2_ThermoPropInqTempCrit()
    m_DensC       (m) = CO2_ThermoPropInqDensCrit()
    ! H2O
    m = m + 1
    m_UnivGasConst(m) = -1.0d100
    m_MolWeight   (m) = -1.0d100
    m_TempC       (m) = -1.0d100
    m_DensC       (m) = -1.0d100
    ! N2
    m = m + 1
    m_UnivGasConst(m) = N2_ThermoPropInqUnivGasConst()
    m_MolWeight   (m) = N2_ThermoPropInqMolWeight()
    m_TempC       (m) = N2_ThermoPropInqTempCrit()
    m_DensC       (m) = N2_ThermoPropInqDensCrit()

    ! mean molecular weight
    MeanMolWeight = 0.0d0
    do m = 1, NComp
      if ( m_Flag(m) ) then
        MeanMolWeight = MeanMolWeight + m_VMR(m) * m_MolWeight(m)
      end if
    end do
    ! gas constant for gas mixture
    GasConst = UnivGasConst / MeanMolWeight

    do m = 1, NComp
      if ( m_Flag(m) ) then
        m_MolDensC         (m) = m_DensC(m) / m_MolWeight(m)
        m_UnivGasConstRatio(m) = m_UnivGasConst(m) / UnivGasConst
      else
        m_MolDensC         (m) = -1.0d100
        m_UnivGasConstRatio(m) = -1.0d100
      end if
    end do

    ! temperature and density used to reduce those values
    TempR = MixTempR()
    DensR = MixDensR()
    MolDensR = DensR / MeanMolWeight


  end subroutine ThermoPropInit

  !----------------------------------------------------------------------------
  ! finalization
  !----------------------------------------------------------------------------

  subroutine ThermoPropFinalization

!!$    deallocate( m_VMR )

  end subroutine ThermoPropFinalization

  !----------------------------------------------------------------------------
  ! inquire molar gas constant
  !----------------------------------------------------------------------------

  function ThermoPropInqUnivGasConst() &
    & result( Val )

    real(dp) :: Val

    Val = UnivGasConst

  end function ThermoPropInqUnivGasConst

  !----------------------------------------------------------------------------
  ! inquire molecular weight
  !----------------------------------------------------------------------------

  function ThermoPropInqMolWeight() &
    & result( Val )

    real(dp) :: Val

    Val = MeanMolWeight

  end function ThermoPropInqMolWeight

  !----------------------------------------------------------------------------
  ! Helmholtz energy
  !----------------------------------------------------------------------------

  function ThermoPropHelmholtzEne( Dens, Temp ) &
    & result( HEne )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: HEne


    ! Local variables
    !
    real(dp) :: Delta
    real(dp) :: Tau
    real(dp) :: HEneIG
    real(dp) :: HEneRes


    Delta = Dens  / DensR
    Tau   = TempR / Temp

    HEneIG  = IGAlpha( Delta, Tau )
    HEneRes = ResAlpha( Delta, Tau )

    if ( FlagIdealGas ) then
      HEne = HEneIG
    else
      HEne = HEneIG + HEneRes
    end if

    HEne = HEne * GasConst * Temp


  end function ThermoPropHelmholtzEne

  !----------------------------------------------------------------------------
  ! compressibility factor  (Table B1 of Kunz and Wagner, 2012)
  !----------------------------------------------------------------------------

  function ThermoPropCompFact( Dens, Temp ) &
    & result( CompFact )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: CompFact


    ! Local variables
    !
    real(dp) :: Delta
    real(dp) :: Tau


    Delta = Dens  / DensR
    Tau   = TempR / Temp

    if ( FlagIdealGas ) then
      CompFact = 1.0d0
    else
      CompFact = 1.0d0 + Delta * ResDAlphaDDelta( Delta, Tau )
    end if


  end function ThermoPropCompFact

  !----------------------------------------------------------------------------
  ! density
  !----------------------------------------------------------------------------

  function ThermoPropDens( Press, Temp ) &
    & result( Dens )

    use thermo_prop_co2_sw1996, only :         &
      & CO2_ThermoPropDens => ThermoPropDens

    real(dp), intent(in) :: Press
    real(dp), intent(in) :: Temp
    real(dp)             :: Dens


    ! Local variables
    !
    character(stdstr) :: CPhase

    real(dp) :: DensCur
    real(dp) :: DensNext
    real(dp) :: DPDRhoCT
    real(dp) :: DelDens

    integer, parameter  :: NLoop = 100
    integer             :: iLoop
    real(dp)            :: a_DensHist( 0:NLoop )

    integer :: m


    ! Initialization
!!$    call ThermoPropJudgePhase( Temp, Press, CPhase )
!!$    select case ( CPhase )
!!$    case ( 'G', 'C' )
!!$      DensCur = Press / ( GasConst * Temp )
!!$    case default
!!$      DensCur = 1500.0d0
!!$    end select
    !
    call ThermoPropJudgePhase( Temp, Press, CPhase )
    DensCur = 0.0d0
    do m = 1, NComp
      if ( m_Flag(m) ) then
        select case ( CPhase(m:m) )
        case ( 'G', 'C' )
          DensCur = DensCur + Press / ( GasConst * Temp ) * m_VMR(m)
        case default
          DensCur = DensCur + 1500.0d0 * m_VMR(m)
        end select
      end if
    end do
    ! CO2 density is used as initial condition
!!$    DensCur = CO2_ThermoPropDens( Press, Temp )

    a_DensHist(0) = DensCur
    search_pressure : do iLoop = 1, NLoop
      DPDRhoCT = ThermoPropDPressDDensConstTemp( DensCur, Temp )
      DelDens =                                                &
        &   (   Press                                          &
        &     - ThermoPropCompFact( DensCur, Temp )   &
        &       * DensCur * GasConst * Temp                  ) &
        & / DPDRhoCT
      DensNext = DensCur + DelDens
      DensNext = max( DensNext, 1.0d-100 )
      a_DensHist( iLoop ) = DensNext
      if ( abs( ( DensNext - DensCur ) / DensCur ) < ConvThreshold ) then
        exit search_pressure
      end if
      DensCur  = DensNext
    end do search_pressure
    if ( iLoop >= NLoop ) then
      write( 6, * ) 'ERROR: In ', trim( ModuleName )
      write( 6, * ) 'ERROR: In ThermoPropDens'
      write( 6, * ) 'ERROR:   Loop does not converge, ', Press, Temp
      write( 6, * ) 'ERROR:   ', iLoop
      write( 6, * ) 'ERROR:   ', a_DensHist
      stop
    end if

    Dens = DensNext


  end function ThermoPropDens

  !----------------------------------------------------------------------------
  ! temperature
  !----------------------------------------------------------------------------

  function ThermoPropTemp( Press, Dens ) &
    & result( Temp )

    real(dp), intent(in) :: Press
    real(dp), intent(in) :: Dens
    real(dp)             :: Temp


    ! Local variables
    !
    real(dp) :: Temp1
    real(dp) :: Temp2
    real(dp) :: DensCur
    real(dp) :: TempCur
    real(dp) :: TempNext
    real(dp) :: DPDTCRho
    real(dp) :: DelTemp
    integer, parameter  :: NLoop = 100
    integer             :: iLoop
    real(dp)            :: a_Hist( 0:NLoop )


    ! Initialization
!!$    if ( Dens < DensC ) then
!!$      ! MEMO
!!$      !   Facot 0.8 is required to obtain appropriate temperatures
!!$      !   when the state is close to liquid-gas boundary.
!!$      !   If this factor is not included, this routine converges to wrong
!!$      !   temperatures in some cases.
!!$      !
!!$      TempCur = Press / ( Dens * GasConst ) / 0.8d0
!!$    else
!!$      TempCur = 200.0d0
!!$      TempCur = 250.0d0
!!$    end if

    ! Bisectional method
    Temp1   = 50.0d0
    DensCur = ThermoPropDens( Press, Temp1 )
    if ( DensCur < Dens ) then
!!$      write( 6, * ) 'ERROR: In ', trim( ModuleName )
      write( 6, * ) 'ERROR: In ThermoPropTemp'
      write( 6, * ) 'ERROR:   Temperature seems to be too low.'
      write( 6, * ) 'ERROR:   ', Dens, DensCur
      stop
    end if
    search_init_temperature : do
      if ( Temp1 < 100.0d0 ) then
        Temp2 = Temp1 + 10.0d0
      else
        Temp2 = Temp1 + 100.0d0
      end if
      DensCur = ThermoPropDens( Press, Temp2 )
      if ( DensCur < Dens ) then
        exit
      end if
      Temp1 = Temp2
    end do search_init_temperature
    bisectional_method : do
      TempCur = ( Temp1 + Temp2 ) / 2.0d0
      DensCur = ThermoPropDens( Press, TempCur )
      if ( DensCur < Dens ) then
        Temp2 = TempCur
      else
        Temp1 = TempCur
      end if
      if ( abs( ( Temp2 - Temp1 ) / Temp1 ) < ConvThreshold ) then
        exit bisectional_method
      end if
    end do bisectional_method
    TempCur = ( Temp1 + Temp2 ) / 2.0d0
    a_Hist(0) = TempCur
    search_temperature : do iLoop = 1, NLoop
      DPDTCRho = ThermoPropDPressDTempConstDens( Dens, TempCur )
      DelTemp =                                                &
        &   (   Press                                          &
        &     - ThermoPropCompFact( Dens, TempCur )   &
        &       * Dens * GasConst * TempCur                  ) &
        & / DPDTCRho
      TempNext = TempCur + DelTemp
      a_Hist( iLoop ) = TempNext
!!$      write( 6, * ) iLoop, TempCur, TempNext
!!$      write( 6, * ) ThermoPropDens( Press, TempNext )
      if ( abs( ( TempNext - TempCur ) / TempCur ) < ConvThreshold ) then
        exit search_temperature
      end if
      TempCur  = TempNext
    end do search_temperature
    if ( iLoop >= NLoop ) then
      write( 6, * ) 'ERROR: In ', trim( ModuleName )
      write( 6, * ) 'ERROR: In ThermoPropTemp'
      write( 6, * ) 'ERROR:   Loop does not converge'
      write( 6, * ) 'ERROR:   ', iLoop
      write( 6, * ) 'ERROR:   ', a_Hist
      stop
    end if

    Temp = TempNext


  end function ThermoPropTemp

  !----------------------------------------------------------------------------
  ! temperature from pressure and enthalpy
  !
  ! This routine calculate temperature by the use of pressure and enthalpy.
  ! But, this routine requires a very small threshold for convergence,
  ! ConvThreshold ~ 1e-9 or 1e-10 or so.
  ! Instead, a routine using ThermoPropDensTempFromPressEnthalpy seems to
  ! calculate better temperatures with larger threshold, ConvThreshold ~ 1e-5.
  !----------------------------------------------------------------------------

  function NG_ThermoPropTempFromPressEnthalpy( Press, Enthalpy ) & ! (in)
    & result( Temp )

    real(dp), intent(in ) :: Press
    real(dp), intent(in ) :: Enthalpy
    real(dp)              :: Temp


    ! Local variables
    !
!!$    real(dp) :: TempMin = 60.0d0
    real(dp) :: Dens
    real(dp) :: TempCur
!!$    real(dp) :: TempNext
    real(dp) :: Temp1
    real(dp) :: Temp2
    real(dp) :: Enthalpy1
    real(dp) :: Enthalpy2
    real(dp) :: EnthalpyCur
!!$    real(dp) :: DHDTCP
!!$    real(dp) :: DelTemp
!!$    integer, parameter  :: NLoop = 100
!!$    integer             :: iLoop
!!$    real(dp)            :: a_HistDens( 0:NLoop )
!!$    real(dp)            :: a_HistTemp( 0:NLoop )


    ! Initialization
    !   find temperature with a given pressure

    ! Bisectional method
    Temp1     = 50.0d0
    Dens      = ThermoPropDens( Press, Temp1 )
    Enthalpy1 = ThermoPropEnthalpy( Dens, Temp1 )
    if ( Enthalpy < Enthalpy1 ) then
      write( 6, * ) 'ERROR: In ', trim( ModuleName )
      write( 6, * ) 'ERROR: In ThermoPropTempFromPressEnthalpy'
      write( 6, * ) 'ERROR:   Temperature seems to be too low.'
      stop
    end if
    search_init_temperature : do
      if ( Temp1 < 100.0d0 ) then
        Temp2     = Temp1 + 10.0d0
      else
        Temp2     = Temp1 + 100.0d0
      end if
      Dens      = ThermoPropDens( Press, Temp2 )
      Enthalpy2 = ThermoPropEnthalpy( Dens, Temp2 )
      if ( Enthalpy < Enthalpy2 ) then
        exit
      end if
      Temp1 = Temp2
    end do search_init_temperature
    bisectional_method : do
      TempCur     = ( Temp1 + Temp2 ) / 2.0d0
      Dens        = ThermoPropDens( Press, TempCur )
      EnthalpyCur = ThermoPropEnthalpy( Dens, TempCur )
      if ( Enthalpy < EnthalpyCur ) then
        Temp2 = TempCur
      else
        Temp1 = TempCur
      end if
      if ( abs( ( Temp2 - Temp1 ) / Temp1 ) < ConvThreshold ) then
        exit bisectional_method
      end if
    end do bisectional_method
    Temp = ( Temp1 + Temp2 ) / 2.0d0


    ! Newton's method
!!$    TempCur = TempMin
!!$    a_HistTemp(0) = TempCur
!!$    search_temperature : do iLoop = 1, NLoop
!!$      Dens        = ThermoPropDens( Press, TempCur )
!!$      EnthalpyCur = ThermoPropEnthalpy( Dens, TempCur )
!!$      DHDTCP      = ThermoPropCp( Dens, TempCur )
!!$      if ( DHDTCP <= 0.0d0 ) then
!!$        write( 6, * ) 'ERROR: In ', trim( ModuleName )
!!$        write( 6, * ) 'ERROR: In ThermoPropTempFromPressEnthalpy'
!!$        write( 6, * ) 'ERROR:   (DHDT)_p is negative, ', DHDTCP
!!$        write( 6, * ) 'ERROR:   at rho = ', Dens, ', T = ', TempCur
!!$        stop
!!$      end if
!!$      DelTemp     = ( Enthalpy - EnthalpyCur ) / DHDTCP
!!$      TempNext    = TempCur + DelTemp
!!$      TempNext    = max( TempNext, TempMin )
!!$      a_HistTemp( iLoop ) = TempNext
!!$      if ( abs( ( TempNext - TempCur ) / TempCur ) < ConvThreshold ) then
!!$        exit search_temperature
!!$      end if
!!$      TempCur = TempNext
!!$    end do search_temperature
!!$    if ( iLoop >= NLoop ) then
!!$      write( 6, * ) 'ERROR: In ', trim( ModuleName )
!!$      write( 6, * ) 'ERROR: In ThermoPropTempFromPressEnthalpy'
!!$      write( 6, * ) 'ERROR:   Loop for temperature does not converge'
!!$      write( 6, * ) 'ERROR:   ', iLoop
!!$      write( 6, * ) 'ERROR:   ', a_HistTemp
!!$      stop
!!$    end if
!!$    Temp = TempNext
!!$
!!$    if ( Temp == TempMin ) then
!!$      write( 6, * ) 'ERROR: In ', trim( ModuleName )
!!$      write( 6, * ) 'ERROR: In ThermoPropTempFromPressEnthalpy'
!!$      write( 6, * ) 'ERROR:   Temperature seems to be calculated, properly.'
!!$      write( 6, * ) 'ERROR:   ', Temp, TempMin
!!$      stop
!!$    end if


  end function NG_ThermoPropTempFromPressEnthalpy

  !----------------------------------------------------------------------------
  ! temperature from pressure and enthalpy
  !----------------------------------------------------------------------------

  function ThermoPropTempFromPressEnthalpy( Press, Enthalpy ) & ! (in)
    & result( Temp )

    real(dp), intent(in ) :: Press
    real(dp), intent(in ) :: Enthalpy
    real(dp)              :: Temp


    ! Local variables
    !
    real(dp) :: Dens


    call ThermoPropDensTempFromPressEnthalpy( &
      & Press, Enthalpy, & ! (in)
      & Dens, Temp       & ! (out)
      & )


  end function ThermoPropTempFromPressEnthalpy

  !----------------------------------------------------------------------------
  ! entropy (Table B1 of Kunz and Wagner, 2012)
  !----------------------------------------------------------------------------

  function ThermoPropEntropy( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: Delta
    real(dp) :: Tau


    Delta = Dens  / DensR
    Tau   = TempR / Temp

    if ( FlagIdealGas ) then
      Val =                           &
        &   Tau * IGDAlphaDTau( Tau ) &
        & - IGAlpha( Delta, Tau )
    else
      Val =                                                             &
        &   Tau * ( IGDAlphaDTau( Tau ) + ResDAlphaDTau( Delta, Tau ) ) &
        & - ( IGAlpha( Delta, Tau ) + ResAlpha( Delta, Tau ) )
    end if

    Val = Val * GasConst


  end function ThermoPropEntropy

  !----------------------------------------------------------------------------
  ! enthalpy (Table B1 of Kunz and Wagner, 2012)
  !----------------------------------------------------------------------------

  function ThermoPropEnthalpy( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: Delta
    real(dp) :: Tau


    Delta = Dens  / DensR
    Tau   = TempR / Temp

    if ( FlagIdealGas ) then
      Val =                                                         &
        &   1.0d0                                                   &
        & + Tau * IGDAlphaDTau( Tau )
    else
      Val =                                                             &
        &   1.0d0                                                       &
        & + Tau * ( IGDAlphaDTau( Tau ) + ResDAlphaDTau( Delta, Tau ) ) &
        & + Delta * ResDAlphaDDelta( Delta, Tau )
    end if

    Val = Val * GasConst * Temp


  end function ThermoPropEnthalpy

  !----------------------------------------------------------------------------
  ! internal energy (Table B1 of Kunz and Wagner, 2012)
  !----------------------------------------------------------------------------

  function ThermoPropInternalEnergy( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: Delta
    real(dp) :: Tau


    Delta = Dens  / DensR
    Tau   = TempR / Temp

    if ( FlagIdealGas ) then
      Val =                                                         &
        &   Tau * IGDAlphaDTau( Tau )
    else
      Val =                                                         &
        & + Tau * ( IGDAlphaDTau( Tau ) + ResDAlphaDTau( Delta, Tau ) )
    end if

    Val = Val * GasConst * Temp


  end function ThermoPropInternalEnergy

  !----------------------------------------------------------------------------
  ! isobaric heat capacity (specific heat on constant pressure)
  ! (Table B1 of Kunz and Wagner, 2012)
  !----------------------------------------------------------------------------

  function ThermoPropCp( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: Delta
    real(dp) :: Tau


    Delta = Dens  / DensR
    Tau   = TempR / Temp

    if ( FlagIdealGas ) then
      Val =                                                           &
        & - Tau**2 * IGD2AlphaDTau2( Tau )                            &
        & + 1.0d0
    else
      Val =                                                             &
        & - Tau**2 * (   IGD2AlphaDTau2( Tau )                          &
        &              + ResD2AlphaDTau2( Delta, Tau ) )                &
        & +   (   1.0d0                                                 &
        &       + Delta * ResDAlphaDDelta( Delta, Tau )                 &
        &       - Delta * Tau * ResD2AlphaDDeltaDTau( Delta, Tau ) )**2 &
        &   / (   1.0d0                                                 &
        &       + 2.0d0 * Delta * ResDAlphaDDelta( Delta, Tau )         &
        &       + Delta**2 * ResD2AlphaDDelta2( Delta, Tau ) )
    end if

    Val = Val * GasConst


  end function ThermoPropCp

  !----------------------------------------------------------------------------
  ! isochoric heat capacity (specific heat on constant volume)
  ! (Table B1 of Kunz and Wagner, 2012)
  !----------------------------------------------------------------------------

  function ThermoPropCv( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: Delta
    real(dp) :: Tau


    Delta = Dens  / DensR
    Tau   = TempR / Temp

    if ( FlagIdealGas ) then
      Val =                                                           &
        & - Tau**2 * IGD2AlphaDTau2( Tau )
    else
      Val =                                                           &
        & - Tau**2 * (   IGD2AlphaDTau2( Tau )                        &
        &              + ResD2AlphaDTau2( Delta, Tau ) )
    end if

    Val = Val * GasConst


  end function ThermoPropCv

  !----------------------------------------------------------------------------
  ! DPDRho
  ! (Table B2 of Kunz and Wagner, 2012)
  !----------------------------------------------------------------------------

  function ThermoPropDPressDDensConstTemp( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: Delta
    real(dp) :: Tau


    Delta = Dens  / DensR
    Tau   = TempR / Temp

    if ( FlagIdealGas ) then
      Val = GasConst * Temp * Delta                         &
        & * (   2.0d0 * IGDAlphaDDelta   ( Delta )          &
        &     + Delta * IGD2AlphaDDelta2 ( Delta ) )
    else
      Val = GasConst * Temp * Delta                             &
        & * (   2.0d0 * (   IGDAlphaDDelta   ( Delta )          &
        &                 + ResDAlphaDDelta  ( Delta, Tau ) )   &
        &     + Delta * (   IGD2AlphaDDelta2 ( Delta )          &
        &                 + ResD2AlphaDDelta2( Delta, Tau ) ) )
    end if


  end function ThermoPropDPressDDensConstTemp

  !----------------------------------------------------------------------------
  ! DPDT
  ! (Table B2 of Kunz and Wagner, 2012)
  !----------------------------------------------------------------------------

  function ThermoPropDPressDTempConstDens( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: Delta
    real(dp) :: Tau


    Delta = Dens / DensR
    Tau   = TempR / Temp

    if ( FlagIdealGas ) then
      Val = Dens * GasConst * Delta                              &
        & * (   IGDAlphaDDelta   ( Delta )                       &
        &     - Tau * IGD2AlphaDDeltaDTau () )
    else
      Val = Dens * GasConst * Delta                              &
        & * (   IGDAlphaDDelta   ( Delta )                       &
        &     + ResDAlphaDDelta  ( Delta, Tau )                  &
        &     - Tau * (   IGD2AlphaDDeltaDTau ()                 &
        &               + ResD2AlphaDDeltaDTau( Delta, Tau ) ) )
    end if


  end function ThermoPropDPressDTempConstDens

  !----------------------------------------------------------------------------
  ! DHDRho
  !----------------------------------------------------------------------------

  function ThermoPropDEnthalpyDDensConstTemp( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: Delta
    real(dp) :: Tau


    Delta = Dens  / DensR
    Tau   = TempR / Temp

    if ( FlagIdealGas ) then
      Val =                                                                   &
        & GasConst * TempR / ( DensR * Tau )                                  &
        & * Tau * IGD2AlphaDDeltaDTau()
    else
      Val =                                                                          &
        & GasConst * TempR / ( DensR * Tau )                                         &
        & * (   Tau * ( IGD2AlphaDDeltaDTau() + ResD2AlphaDDeltaDTau( Delta, Tau ) ) &
        &     + ResDAlphaDDelta( Delta, Tau ) + Delta * ResD2AlphaDDelta2( Delta, Tau ) )
    end if


  end function ThermoPropDEnthalpyDDensConstTemp

  !----------------------------------------------------------------------------
  ! DHDT
  !----------------------------------------------------------------------------

  function ThermoPropDEnthalpyDTempConstDens( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: Delta
    real(dp) :: Tau


    Delta = Dens  / DensR
    Tau   = TempR / Temp

    if ( FlagIdealGas ) then
      Val =                                                       &
        & GasConst                                                &
        & * (   1.0d0                                             &
        &     - Tau**2 * IGD2AlphaDTau2( Tau ) )
    else
      Val =                                                                        &
        & GasConst                                                                 &
        & * (   1.0d0                                                              &
        &     + Delta * ResDAlphaDDelta( Delta, Tau )                              &
        &     - Tau**2 * ( IGD2AlphaDTau2( Tau ) + ResD2AlphaDTau2( Delta, Tau ) ) &
        &     - Delta * Tau * ResD2AlphaDDeltaDTau( Delta, Tau ) )
    end if


  end function ThermoPropDEnthalpyDTempConstDens

  !----------------------------------------------------------------------------
  ! thermal expansion coefficient
  !----------------------------------------------------------------------------

  function ThermoPropThermoExpCoef( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !


    Val = &
      &   ThermoPropDPressDTempConstDens( Dens, Temp ) &
      & / ( ThermoPropDPressDDensConstTemp( Dens, Temp ) * Dens )


  end function ThermoPropThermoExpCoef

  !----------------------------------------------------------------------------
  ! adiabatic lapse rate
  !----------------------------------------------------------------------------

  function ThermoPropAdLapRateDTDP( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: Cp
    real(dp) :: ExpCoef


    Cp       = ThermoPropCp           ( Dens, Temp )
    ExpCoef  = ThermoPropThermoExpCoef( Dens, Temp )

    Val = Temp / ( Cp * Dens ) * ExpCoef


  end function ThermoPropAdLapRateDTDP

  !----------------------------------------------------------------------------
  ! temperature and density from pressure and enthalpy
  !----------------------------------------------------------------------------

  subroutine ThermoPropDensTempFromPressEnthalpy( &
    & Press, Enthalpy, & ! (in)
    & Dens, Temp       & ! (out)
    & )

    real(dp), intent(in ) :: Press
    real(dp), intent(in ) :: Enthalpy
    real(dp), intent(out) :: Dens
    real(dp), intent(out) :: Temp


    ! Local variables
    !
    real(dp) :: Temp1
    real(dp) :: Temp2
!!$    real(dp) :: Dens1
!!$    real(dp) :: Dens2
    real(dp) :: Enthalpy1
    real(dp) :: Enthalpy2

    real(dp) :: Delta
    real(dp) :: Tau
    real(dp) :: DensCur
    real(dp) :: DensNext
    real(dp) :: TempCur
    real(dp) :: TempNext
    real(dp) :: PressCur
    real(dp) :: EnthalpyCur
    real(dp) :: DPDRhoCT
    real(dp) :: DPDTCRho
    real(dp) :: DHDRhoCT
    real(dp) :: DHDTCRho
    real(dp) :: DelDens
    real(dp) :: DelTemp
    integer, parameter  :: NLoop = 100
    integer             :: iLoop
    real(dp)            :: a_HistDens( 0:NLoop )
    real(dp)            :: a_HistTemp( 0:NLoop )


    ! Initialization
    !   find temperature with a given pressure

    ! Bisectional method
    Temp1     = 50.0d0
    DensCur   = ThermoPropDens( Press, Temp1 )
    Enthalpy1 = ThermoPropEnthalpy( DensCur, Temp1 )
    if ( Enthalpy < Enthalpy1 ) then
      write( 6, * ) 'ERROR: In ', trim( ModuleName )
      write( 6, * ) 'ERROR: In ThermoPropDensTempFromPressEnthalpy'
      write( 6, * ) 'ERROR:   Temperature seems to be too low.'
      stop
    end if
    search_init_temperature : do
      if ( Temp1 < 100.0d0 ) then
        Temp2     = Temp1 + 10.0d0
      else
        Temp2     = Temp1 + 100.0d0
      end if
      DensCur   = ThermoPropDens( Press, Temp2 )
      Enthalpy2 = ThermoPropEnthalpy( DensCur, Temp2 )
      if ( Enthalpy < Enthalpy2 ) then
        exit
      end if
      Temp1 = Temp2
    end do search_init_temperature
    bisectional_method : do
      TempCur     = ( Temp1 + Temp2 ) / 2.0d0
      DensCur     = ThermoPropDens( Press, TempCur )
      EnthalpyCur = ThermoPropEnthalpy( DensCur, TempCur )
      if ( Enthalpy < EnthalpyCur ) then
        Temp2 = TempCur
      else
        Temp1 = TempCur
      end if
      if ( abs( ( Temp2 - Temp1 ) / Temp1 ) < ConvThreshold ) then
        exit bisectional_method
      end if
    end do bisectional_method
    TempCur = ( Temp1 + Temp2 ) / 2.0d0


    ! Newton's method
    ! NOTE: This does not work well in some cases, i.e., case of liquid.
!!$    TempCur = 300.0d0
!!$    a_HistTemp(0) = TempCur
!!$    search_temperature : do iLoop = 1, NLoop
!!$      Temp1     = TempCur
!!$      Temp2     = Temp1 + 1.0d-1
!!$      Dens1     = ThermoPropDens( Press, Temp1 )
!!$      Dens2     = ThermoPropDens( Press, Temp2 )
!!$      Enthalpy1 = ThermoPropEnthalpy( Dens1, Temp1 )
!!$      Enthalpy2 = ThermoPropEnthalpy( Dens2, Temp2 )
!!$      ! (DHDT)_p
!!$      DHDT = ( Enthalpy2 - Enthalpy1 ) / ( Temp2 - Temp1 )
!!$      !
!!$      EnthalpyCur = Enthalpy1
!!$      DelTemp = ( Enthalpy - EnthalpyCur ) / DHDT
!!$      TempNext = TempCur + DelTemp
!!$      TempNext = max( TempNext, 50.0d0 )
!!$      a_HistTemp( iLoop ) = TempNext
!!$      write( 6, * ) iLoop, TempCur, TempNext
!!$      if ( abs( ( TempNext - TempCur ) / TempCur ) < ConvThreshold ) then
!!$        exit search_temperature
!!$      end if
!!$      TempCur = TempNext
!!$    end do search_temperature
!!$    if ( iLoop >= NLoop ) then
!!$      write( 6, * ) 'ERROR: In ', trim( ModuleName )
!!$      write( 6, * ) 'ERROR: In ThermoPropDensTempFromPressEnthalpy'
!!$      write( 6, * ) 'ERROR:   Loop for temperature does not converge'
!!$      write( 6, * ) 'ERROR:   ', iLoop
!!$      write( 6, * ) 'ERROR:   ', a_HistTemp
!!$      stop
!!$    end if
!!$    TempCur = TempNext


    DensCur = ThermoPropDens( Press, TempCur )

    a_HistDens(0) = DensCur
    a_HistTemp(0) = TempCur
    search_density_temperature : do iLoop = 1, NLoop
      Delta = DensCur / DensR
      Tau   = TempR / TempCur

      PressCur    =                                       &
        & ThermoPropCompFact( DensCur, TempCur ) &
        & * DensCur * GasConst * TempCur
      EnthalpyCur = ThermoPropEnthalpy( DensCur, TempCur )

      DPDRhoCT = ThermoPropDPressDDensConstTemp( DensCur, TempCur )
      DPDTCRho = ThermoPropDPressDTempConstDens( DensCur, TempCur )

      if ( FlagIdealGas ) then
!!$        DHDRhoCT =                                                              &
!!$          & GasConst * TempR / ( DensR * Tau )                                &
!!$          & * Tau * IGD2AlphaDDeltaDTau()
        DHDRhoCT = 0.0d0
        DHDTCRho =                                                  &
          & GasConst                                                &
          & * (   1.0d0                                             &
          &     - Tau**2 * IGD2AlphaDTau2( Tau ) )
        DelTemp =                                         &
          &   (   ( Enthalpy - EnthalpyCur ) * DPDRhoCT   &
          &     - ( Press    - PressCur    ) * DHDRhoCT ) &
          & / ( DHDTCRho * DPDRhoCT - DHDRhoCT * DPDTCRho )
        TempNext = TempCur + DelTemp
        DelDens = Press / ( GasConst * TempNext ) - DensCur
        DensNext = DensCur + DelDens
      else
        DHDRhoCT =                                                            &
          & GasConst * TempR / ( DensR * Tau )                                &
          & * (   Tau * ( IGD2AlphaDDeltaDTau() + ResD2AlphaDDeltaDTau( Delta, Tau ) ) &
          &     + ResDAlphaDDelta( Delta, Tau ) + Delta * ResD2AlphaDDelta2( Delta, Tau ) )
        DHDTCRho =                                                  &
          & GasConst                                                &
          & * (   1.0d0                                             &
          &     + Delta * ResDAlphaDDelta( Delta, Tau )             &
          &     - Tau**2 * ( IGD2AlphaDTau2( Tau ) + ResD2AlphaDTau2( Delta, Tau ) ) &
          &     - Delta * Tau * ResD2AlphaDDeltaDTau( Delta, Tau ) )
        DelTemp =                                         &
          &   (   ( Enthalpy - EnthalpyCur ) * DPDRhoCT   &
          &     - ( Press    - PressCur    ) * DHDRhoCT ) &
          & / ( DHDTCRho * DPDRhoCT - DHDRhoCT * DPDTCRho )
        DelDens = &
          & ( Enthalpy - EnthalpyCur - DHDTCRho * DelTemp ) / DHDRhoCT
        DensNext = DensCur + DelDens
        TempNext = TempCur + DelTemp
      end if

      a_HistDens( iLoop ) = DensNext
      a_HistTemp( iLoop ) = TempNext
!!$      write( 6, * ) iLoop, DensCur, DensNext, TempCur, TempNext
      if ( ( abs( ( DensNext - DensCur ) / DensCur ) < ConvThreshold ) .and. &
        &  ( abs( ( TempNext - TempCur ) / TempCur ) < ConvThreshold ) ) then
        exit search_density_temperature
      end if
      DensCur  = DensNext
      TempCur  = TempNext
    end do search_density_temperature
    if ( iLoop >= NLoop ) then
      write( 6, * ) 'ERROR: In ', trim( ModuleName )
      write( 6, * ) 'ERROR: In ThermoPropDensTempFromPressEnthalpy'
      write( 6, * ) 'ERROR:   Loop does not converge'
      write( 6, * ) 'ERROR:   ', iLoop
      write( 6, * ) 'ERROR:   ', a_HistDens
      write( 6, * ) 'ERROR:   ', a_HistTemp
      stop
    end if

    Dens = DensNext
    Temp = TempNext


  end subroutine ThermoPropDensTempFromPressEnthalpy

  !----------------------------------------------------------------------------
  ! judge phase
  !----------------------------------------------------------------------------

  subroutine ThermoPropJudgePhase( Temp, Press, CPhase )

    use thermo_prop_co2_sw1996, only :                   &
      & CO2_ThermoPropJudgePhase => ThermoPropJudgePhase
    use thermo_prop_n2_s2000, only :                     &
      & N2_ThermoPropJudgePhase  => ThermoPropJudgePhase

    real(dp)    , intent(in ) :: Temp
    real(dp)    , intent(in ) :: Press
    character(*), intent(out) :: CPhase


    ! Local variables
    !
    character(stdstr) :: CPhaseEach
    integer           :: m


    m = 0
    CPhase = ''

    ! CO2
    m = m + 1
    call CO2_ThermoPropJudgePhase( Temp, Press, CPhaseEach )
    CPhase(m:m) = CPhaseEach(1:1)

    ! H2O
    m = m + 1
    CPhase(m:m) = '-'

    ! N2
    m = m + 1
    call N2_ThermoPropJudgePhase( Temp, Press, CPhaseEach )
    CPhase(m:m) = CPhaseEach(1:1)


  end subroutine ThermoPropJudgePhase

  !----------------------------------------------------------------------------
  ! judge phase
  !  If all phases are gas or super-critical, .true. is returned.
  !----------------------------------------------------------------------------

  function ThermoPropJudgePhaseGC( Temp, Press ) &
    & result( Flag )

    real(dp)    , intent(in ) :: Temp
    real(dp)    , intent(in ) :: Press
    logical                   :: Flag


    ! Local variables
    !
    character(stdstr) :: CPhase
    integer           :: m


    call ThermoPropJudgePhase( Temp, Press, CPhase )

    Flag = .true.
    do m = 1, NComp
      if ( m_Flag(m) ) then
        if ( ( CPhase(m:m) /= 'G' ) .and. ( CPhase(m:m) /= 'C' ) ) then
          Flag = .false.
          exit
        end if
      end if
    end do


  end function ThermoPropJudgePhaseGC

  !============================================================================
  ! Internal routines
  !============================================================================

  !----------------------------------------------------------------------------
  ! reducing density for calculations of residual part
  !----------------------------------------------------------------------------

  function MixDensR() &
    & result( Val )

    real(dp) :: Val


    ! Local variables
    !
    integer  :: i
    integer  :: j


    Val = 0.0d0

    do i = 1, NComp
      if ( m_Flag(i) ) then
        Val = Val + m_VMR(i)**2 / m_MolDensC(i)
      end if
    end do
    do i = 1, NComp-1
      if ( .not. m_Flag(i) ) cycle
      do j = i+1, NComp
        if ( m_Flag(j) ) then
          Val = Val &
            & + 2.0d0 * m_VMR(i) * m_VMR(j) &
            &   * mm_ResBetaV(i,j) * mm_ResGammaV(i,j) &
            &   * ( m_VMR(i) + m_VMR(j) ) &
            &   / ( mm_ResBetaV(i,j)**2 * m_VMR(i) + m_VMR(j) ) &
            &   * (   1.0d0 / m_MolDensC(i)**(1.0d0/3.0d0)      &
            &       + 1.0d0 / m_MolDensC(j)**(1.0d0/3.0d0) )**3 &
            &   / 8.0d0
        end if
      end do
    end do
    Val = 1.0d0 / Val

    Val = Val * MeanMolWeight


  end function MixDensR

  !----------------------------------------------------------------------------
  ! reducing temperature for calculations of residual part
  !----------------------------------------------------------------------------

  function MixTempR() &
    & result( Val )

    real(dp) :: Val


    ! Local variables
    !
    integer  :: i
    integer  :: j


    Val = 0.0d0

    do i = 1, NComp
      if ( m_Flag(i) ) then
        Val = Val + m_VMR(i)**2 * m_TempC(i)
      end if
    end do
    do i = 1, NComp-1
      if ( .not. m_Flag(i) ) cycle
      do j = i+1, NComp
        if ( m_Flag(j) ) then
          Val = Val &
            & + 2.0d0 * m_VMR(i) * m_VMR(j) &
            &   * mm_ResBetaT(i,j) * mm_ResGammaT(i,j) &
            &   * ( m_VMR(i) + m_VMR(j) ) &
            &   / ( mm_ResBetaT(i,j)**2 * m_VMR(i) + m_VMR(j) ) &
            &   * sqrt( m_TempC(i) * m_TempC(j) )
        end if
      end do
    end do


  end function MixTempR

  !----------------------------------------------------------------------------
  ! ideal gas part of the reduced Helmholtz energy
  !----------------------------------------------------------------------------

  function IGAlpha( Delta, Tau ) &
    & result( HEne )

    use thermo_prop_co2_sw1996, only :                     &
      & CO2_IGAlpha              => IGPhi
    use thermo_prop_n2_s2000, only :                       &
      & N2_IGAlpha               => IGAlpha

    real(dp), intent(in) :: Delta
    real(dp), intent(in) :: Tau
    real(dp)             :: HEne


    ! Local variables
    !
    real(dp) :: MolDens
    real(dp) :: Temp
    real(dp) :: DeltaE
    real(dp) :: TauE
    integer  :: m


    MolDens = Delta * DensR / MeanMolWeight
    Temp    = TempR / Tau


    m = 0
    HEne = 0.0d0

    ! CO2
    m = m + 1
    DeltaE = MolDens / m_MolDensC(m)
    TauE   = m_TempC(m) / Temp
    HEne = HEne &
      & + m_VMR(m) * CO2_IGAlpha( DeltaE, TauE ) * m_UnivGasConstRatio(m)
    if ( FlagMixingEffect ) then
      HEne = HEne &
        & + m_VMR(m) * log( m_VMR(m) )
    end if

    ! H2O
    m = m + 1
    ! NOT SUPPORTED

    ! N2
    m = m + 1
    DeltaE = MolDens / m_MolDensC(m)
    TauE   = m_TempC(m) / Temp
    HEne = HEne &
      & + m_VMR(m) * N2_IGAlpha( DeltaE, TauE ) * m_UnivGasConstRatio(m)
    if ( FlagMixingEffect ) then
      HEne = HEne &
        & + m_VMR(m) * log( m_VMR(m) )
    end if


  end function IGAlpha

  !----------------------------------------------------------------------------
  ! delta derivative of ideal gas part of the reduced Helmholtz energy
  ! Table B5 of Kunz and Wagner (2012)
  !----------------------------------------------------------------------------

  function IGDAlphaDDelta( Delta ) &
    & result( Val )

    use thermo_prop_co2_sw1996, only :                      &
      & CO2_IGDAlphaDDelta       => IGDPhiDDelta
    use thermo_prop_n2_s2000, only :                       &
      & N2_IGDAlphaDDelta        => IGDAlphaDDelta

    real(dp), intent(in) :: Delta
    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: MolDens
!!$    real(dp) :: Temp
    real(dp) :: DeltaE
!!$    real(dp) :: TauE
    integer  :: m


    MolDens = Delta * DensR / MeanMolWeight
!!$    Temp    = TempR / Tau

    Val = 0.0d0

    m = 0

    ! CO2
    m = m + 1
    DeltaE = MolDens / m_MolDensC(m)
!!$    TauE   = m_TempC(m) / Temp
    Val = Val &
      & + m_VMR(m) * MolDensR / m_MolDensC(m) &
      &   * CO2_IGDAlphaDDelta( DeltaE ) * m_UnivGasConstRatio(m)

    ! H2O
    m = m + 1
    ! NOT SUPPORTED

    ! N2
    m = m + 1
    DeltaE = MolDens / m_MolDensC(m)
!!$    TauE   = m_TempC(m) / Temp
    Val = Val &
      & + m_VMR(m) * MolDensR / m_MolDensC(m) &
      &   * N2_IGDAlphaDDelta( DeltaE ) * m_UnivGasConstRatio(m)


  end function IGDAlphaDDelta

  !----------------------------------------------------------------------------
  ! delta second derivative of ideal gas part of the reduced Helmholtz energy
  ! Table B5 of Kunz and Wagner (2012)
  !----------------------------------------------------------------------------

  function IGD2AlphaDDelta2( Delta ) &
    & result( Val )

    use thermo_prop_co2_sw1996, only :                      &
      & CO2_IGD2AlphaDDelta2     => IGD2PhiDDelta2
    use thermo_prop_n2_s2000, only :                       &
      & N2_IGD2AlphaDDelta2      => IGD2AlphaDDelta2

    real(dp), intent(in) :: Delta
    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: MolDens
!!$    real(dp) :: Temp
    real(dp) :: DeltaE
!!$    real(dp) :: TauE
    integer  :: m


    MolDens = Delta * DensR / MeanMolWeight
!!$    Temp    = TempR / Tau


    Val = 0.0d0

    m = 0

    ! CO2
    m = m + 1
    DeltaE = MolDens / m_MolDensC(m)
!!$    TauE   = m_TempC(m) / Temp
    Val = Val &
      & + m_VMR(m) * ( MolDensR / m_MolDensC(m) )**2 &
      &   * CO2_IGD2AlphaDDelta2( DeltaE ) * m_UnivGasConstRatio(m)

    ! H2O
    m = m + 1
    ! NOT SUPPORTED

    ! N2
    m = m + 1
    DeltaE = MolDens / m_MolDensC(m)
!!$    TauE   = m_TempC(m) / Temp
    Val = Val &
      & + m_VMR(m) * ( MolDensR / m_MolDensC(m) )**2 &
      &   * N2_IGD2AlphaDDelta2( DeltaE ) * m_UnivGasConstRatio(m)


  end function IGD2AlphaDDelta2

  !----------------------------------------------------------------------------
  ! tau derivative of ideal gas part of the reduced Helmholtz energy
  ! Table B5 of Kunz and Wagner (2012)
  !----------------------------------------------------------------------------

  function IGDAlphaDTau( Tau ) &
    & result( Val )

    use thermo_prop_co2_sw1996, only :                     &
      & CO2_IGDAlphaDTau         => IGDPhiDTau
    use thermo_prop_n2_s2000, only :                       &
      & N2_IGDAlphaDTau          => IGDAlphaDTau

    real(dp), intent(in) :: Tau
    real(dp)             :: Val


    ! Local variables
    !
!!$    real(dp) :: MolDens
    real(dp) :: Temp
!!$    real(dp) :: DeltaE
    real(dp) :: TauE
    integer  :: m


!!$    MolDens = Delta * DensR / MeanMolWeight
    Temp    = TempR / Tau

    Val = 0.0d0

    m = 0

    ! CO2
    m = m + 1
!!$    DeltaE = MolDens / m_MolDensC(m)
    TauE   = m_TempC(m) / Temp
    Val = Val &
      & + m_VMR(m) * m_TempC(m) / TempR &
      &   * CO2_IGDAlphaDTau( TauE ) * m_UnivGasConstRatio(m)

    ! H2O
    m = m + 1
    ! NOT SUPPORTED

    ! N2
    m = m + 1
!!$    DeltaE = MolDens / m_MolDensC(m)
    TauE   = m_TempC(m) / Temp
    Val = Val &
      & + m_VMR(m) * m_TempC(m) / TempR &
      &   * N2_IGDAlphaDTau( TauE ) * m_UnivGasConstRatio(m)


  end function IGDAlphaDTau

  !----------------------------------------------------------------------------
  ! tau second derivative of ideal gas part of the reduced Helmholtz energy
  ! Table B5 of Kunz and Wagner (2012)
  !----------------------------------------------------------------------------

  function IGD2AlphaDTau2( Tau ) &
    & result( Val )

    use thermo_prop_co2_sw1996, only :                     &
      & CO2_IGD2AlphaDTau2       => IGD2PhiDTau2
    use thermo_prop_n2_s2000, only :                       &
      & N2_IGD2AlphaDTau2        => IGD2AlphaDTau2

    real(dp), intent(in) :: Tau
    real(dp)             :: Val


    ! Local variables
    !
!!$    real(dp) :: MolDens
    real(dp) :: Temp
!!$    real(dp) :: DeltaE
    real(dp) :: TauE
    integer  :: m


!!$    MolDens = Delta * DensR / MeanMolWeight
    Temp    = TempR / Tau

    Val = 0.0d0

    m = 0

    ! CO2
    m = m + 1
!!$    DeltaE = MolDens / m_MolDensC(m)
    TauE   = m_TempC(m) / Temp
    Val = Val &
      & + m_VMR(m) * ( m_TempC(m) / TempR )**2 &
      &   * CO2_IGD2AlphaDTau2( TauE ) * m_UnivGasConstRatio(m)

    ! H2O
    m = m + 1
    ! NOT SUPPORTED

    ! N2
    m = m + 1
!!$    DeltaE = MolDens / m_MolDensC(m)
    TauE   = m_TempC(m) / Temp
    Val = Val &
      & + m_VMR(m) * ( m_TempC(m) / TempR )**2 &
      &   * N2_IGD2AlphaDTau2( TauE ) * m_UnivGasConstRatio(m)


  end function IGD2AlphaDTau2

  !----------------------------------------------------------------------------
  ! delta and tau derivative of ideal gas part of the reduced Helmholtz energy
  ! Table B5 of Kunz and Wagner (2012)
  !----------------------------------------------------------------------------

  function IGD2AlphaDDeltaDTau( ) &
    & result( Val )

    use thermo_prop_co2_sw1996, only :                      &
      & CO2_IGD2AlphaDDeltaDTau  => IGD2PhiDDeltaDTau
    use thermo_prop_n2_s2000, only :                       &
      & N2_IGD2AlphaDDeltaDTau   => IGD2AlphaDDeltaDTau

!!$    real(dp), intent(in) :: Delta
!!$    real(dp), intent(in) :: Tau
    real(dp)             :: Val


    ! Local variables
    !
!!$    real(dp) :: MolDens
!!$    real(dp) :: Temp
!!$    real(dp) :: DeltaE
!!$    real(dp) :: TauE
    integer  :: m


!!$    MolDens = Delta * DensR / MeanMolWeight
!!$    Temp    = TempR / Tau


    Val = 0.0d0

    m = 0

    ! CO2
    m = m + 1
!!$    DeltaE = MolDens / m_MolDensC(m)
!!$    TauE   = m_TempC(m) / Temp
    Val = Val &
      & + m_VMR(m) * ( MolDensR / m_MolDensC(m) ) * ( m_TempC(m) / TempR ) &
      &   * CO2_IGD2AlphaDDeltaDTau() * m_UnivGasConstRatio(m)

    ! H2O
    m = m + 1
    ! NOT SUPPORTED

    ! N2
    m = m + 1
!!$    DeltaE = MolDens / m_MolDensC(m)
!!$    TauE   = m_TempC(m) / Temp
    Val = Val &
      & + m_VMR(m) * ( MolDensR / m_MolDensC(m) ) * ( m_TempC(m) / TempR ) &
      &   * N2_IGD2AlphaDDeltaDTau() * m_UnivGasConstRatio(m)


  end function IGD2AlphaDDeltaDTau

  !----------------------------------------------------------------------------
  ! residual part of the reduced Helmholtz energy
  !----------------------------------------------------------------------------

  function ResAlpha( Delta, Tau ) &
    & result( HEne )

    real(dp), intent(in) :: Delta
    real(dp), intent(in) :: Tau
    real(dp)             :: HEne


    ! Local variables
    !


    HEne = ResAlphaFuncs( "1", Delta, Tau )


  end function ResAlpha

  !----------------------------------------------------------------------------
  ! delta derivative of residual part of the reduced Helmholtz energy
  !----------------------------------------------------------------------------

  function ResDAlphaDDelta( Delta, Tau ) &
    & result( HEne )

    real(dp), intent(in) :: Delta
    real(dp), intent(in) :: Tau
    real(dp)             :: HEne


    ! Local variables
    !


    HEne = ResAlphaFuncs( "DDDelta", Delta, Tau )


  end function ResDAlphaDDelta

  !----------------------------------------------------------------------------
  ! delta second derivative of residual part of the reduced Helmholtz energy
  !----------------------------------------------------------------------------

  function ResD2AlphaDDelta2( Delta, Tau ) &
    & result( HEne )

    real(dp), intent(in) :: Delta
    real(dp), intent(in) :: Tau
    real(dp)             :: HEne


    ! Local variables
    !


    HEne = ResAlphaFuncs( "D2DDelta2", Delta, Tau )


  end function ResD2AlphaDDelta2

  !----------------------------------------------------------------------------
  ! tau derivative of residual part of the reduced Helmholtz energy
  !----------------------------------------------------------------------------

  function ResDAlphaDTau( Delta, Tau ) &
    & result( HEne )

    real(dp), intent(in) :: Delta
    real(dp), intent(in) :: Tau
    real(dp)             :: HEne


    ! Local variables
    !


    HEne = ResAlphaFuncs( "DDTau", Delta, Tau )


  end function ResDAlphaDTau

  !----------------------------------------------------------------------------
  ! tau second derivative of residual part of the reduced Helmholtz energy
  !----------------------------------------------------------------------------

  function ResD2AlphaDTau2( Delta, Tau ) &
    & result( HEne )

    real(dp), intent(in) :: Delta
    real(dp), intent(in) :: Tau
    real(dp)             :: HEne


    ! Local variables
    !


    HEne = ResAlphaFuncs( "D2DTau2", Delta, Tau )


  end function ResD2AlphaDTau2

  !----------------------------------------------------------------------------
  ! delta and tau derivative of residual part of the reduced Helmholtz energy
  !----------------------------------------------------------------------------

  function ResD2AlphaDDeltaDTau( Delta, Tau ) &
    & result( HEne )

    real(dp), intent(in) :: Delta
    real(dp), intent(in) :: Tau
    real(dp)             :: HEne


    ! Local variables
    !


    HEne = ResAlphaFuncs( "D2DDeltaDTau", Delta, Tau )


  end function ResD2AlphaDDeltaDTau

  !----------------------------------------------------------------------------
  ! residual part of the reduced Helmholtz energy and its derivatives
  !----------------------------------------------------------------------------

  function ResAlphaFuncs( Symbol, Delta, Tau ) &
    & result( HEne )

    use thermo_prop_co2_sw1996, only :                      &
      & CO2_ResAlpha              => ResPhi,                &
      & CO2_ResDAlphaDDelta       => ResDPhiDDelta,         &
      & CO2_ResD2AlphaDDelta2     => ResD2PhiDDelta2,       &
      & CO2_ResDAlphaDTau         => ResDPhiDTau,           &
      & CO2_ResD2AlphaDTau2       => ResD2PhiDTau2,         &
      & CO2_ResD2AlphaDDeltaDTau  => ResD2PhiDDeltaDTau
    use thermo_prop_n2_s2000, only :                       &
      & N2_ResAlpha              => ResAlpha,              &
      & N2_ResDAlphaDDelta       => ResDAlphaDDelta,       &
      & N2_ResD2AlphaDDelta2     => ResD2AlphaDDelta2,     &
      & N2_ResDAlphaDTau         => ResDAlphaDTau,         &
      & N2_ResD2AlphaDTau2       => ResD2AlphaDTau2,       &
      & N2_ResD2AlphaDDeltaDTau  => ResD2AlphaDDeltaDTau

    character(*), intent(in) :: Symbol
    real(dp)    , intent(in) :: Delta
    real(dp)    , intent(in) :: Tau
    real(dp)                 :: HEne


    ! Local variables
    !
    real(dp) :: AlphaFunc
    integer  :: m


    m = 0
    HEne = 0.0d0

    ! CO2
    m = m + 1
    select case ( trim( Symbol ) )
    case ( "1" )
      AlphaFunc = CO2_ResAlpha( Delta, Tau )
    case ( "DDDelta" )
      AlphaFunc = CO2_ResDAlphaDDelta( Delta, Tau )
    case ( "D2DDelta2" )
      AlphaFunc = CO2_ResD2AlphaDDelta2( Delta, Tau )
    case ( "DDTau" )
      AlphaFunc = CO2_ResDAlphaDTau( Delta, Tau )
    case ( "D2DTau2" )
      AlphaFunc = CO2_ResD2AlphaDTau2( Delta, Tau )
    case ( "D2DDeltaDTau" )
      AlphaFunc = CO2_ResD2AlphaDDeltaDTau( Delta, Tau )
    case default
      write( 6, * ) 'ERROR: In ', trim( ModuleName )
      write( 6, * ) 'ERROR: In ResAlphaFuncs, CO2'
      write( 6, * ) 'ERROR: Unexpected Symbol, ', trim( Symbol )
      stop
    end select
    AlphaFunc = AlphaFunc * m_UnivGasConstRatio(m)
    HEne = HEne + m_VMR(m) * AlphaFunc

    ! H2O
    m = m + 1
    ! NOT SUPPORTED

    ! N2
    m = m + 1
    select case ( trim( Symbol ) )
    case ( "1" )
      AlphaFunc = N2_ResAlpha( Delta, Tau )
    case ( "DDDelta" )
      AlphaFunc = N2_ResDAlphaDDelta( Delta, Tau )
    case ( "D2DDelta2" )
      AlphaFunc = N2_ResD2AlphaDDelta2( Delta, Tau )
    case ( "DDTau" )
      AlphaFunc = N2_ResDAlphaDTau( Delta, Tau )
    case ( "D2DTau2" )
      AlphaFunc = N2_ResD2AlphaDTau2( Delta, Tau )
    case ( "D2DDeltaDTau" )
      AlphaFunc = N2_ResD2AlphaDDeltaDTau( Delta, Tau )
    case default
      write( 6, * ) 'ERROR: In ', trim( ModuleName )
      write( 6, * ) 'ERROR: In ResAlphaFuncs, N2'
      write( 6, * ) 'ERROR: Unexpected Symbol, ', trim( Symbol )
      stop
    end select
    AlphaFunc = AlphaFunc * m_UnivGasConstRatio(m)
    HEne = HEne + m_VMR(m) * AlphaFunc


    ! departure function
    if ( FlagMixingEffect ) then
      AlphaFunc = DepFuncs( Symbol, Delta, Tau )
      HEne = HEne + AlphaFunc
    end if


  end function ResAlphaFuncs

  !----------------------------------------------------------------------------
  ! departure function (Eq. (6)) and its derivative
  !----------------------------------------------------------------------------

  function DepFuncs( Symbol, Delta, Tau ) &
    & result( Val )

    character(*), intent(in) :: Symbol
    real(dp)    , intent(in) :: Delta
    real(dp)    , intent(in) :: Tau
    real(dp)                 :: Val


    ! Local variables
    !
    real(dp) :: BSDFValue
    integer  :: i
    integer  :: j


    Val = 0.0d0

    do i = 1, NComp-1
      if ( .not. m_Flag(i) ) cycle
      do j = i+1, NComp
        if ( m_Flag(j) ) then
          select case ( trim( Symbol ) )
          case ( "1" )
            BSDFValue = BSDF            ( Delta, Tau, i, j )
          case ( "DDDelta" )
            BSDFValue = DBSDFDDelta     ( Delta, Tau, i, j )
          case ( "D2DDelta2" )
            BSDFValue = D2BSDFDDelta2   ( Delta, Tau, i, j )
          case ( "DDTau" )
            BSDFValue = DBSDFDTau       ( Delta, Tau, i, j )
          case ( "D2DTau2" )
            BSDFValue = D2BSDFDTau2     ( Delta, Tau, i, j )
          case ( "D2DDeltaDTau" )
            BSDFValue = D2BSDFDDeltaDTau( Delta, Tau, i, j )
          case default
            write( 6, * ) 'ERROR: In ', trim( ModuleName )
            write( 6, * ) 'ERROR: In DepFunc'
            write( 6, * ) 'ERROR: Unexpected Symbol, ', trim( Symbol )
            stop
          end select
          Val = Val &
            & + m_VMR(i) * m_VMR(j) &
            &   * mm_ResF(i,j) * BSDFValue
        end if
      end do
    end do


  end function DepFuncs

  !----------------------------------------------------------------------------
  ! binary specific departure function (BSDF) (Eq. (9))
  !----------------------------------------------------------------------------

  function BSDF( Delta, Tau, i, j ) &
    & result( Val )

    real(dp), intent(in) :: Delta
    real(dp), intent(in) :: Tau
    integer , intent(in) :: i
    integer , intent(in) :: j
    real(dp)             :: Val


    ! Local variables
    !
    integer  :: k


    Val = 0.0d0

    do k = 1, mm_KPol(i,j)
      Val = Val &
        & + amm_BinDepFuncSmallN(k,i,j) &
        &   * Delta**amm_BinDepFuncSmallD(k,i,j) &
        &   * Tau**amm_BinDepFuncSmallT(k,i,j)
    end do
    do k = mm_KPol(i,j)+1, mm_KPol(i,j)+mm_KExp(i,j)
      Val = Val &
        & + amm_BinDepFuncSmallN(k,i,j) &
        &   * Delta**amm_BinDepFuncSmallD(k,i,j) &
        &   * Tau**amm_BinDepFuncSmallT(k,i,j) &
        &   * exp( -Delta**amm_BinDepFuncSmallL(k,i,j) )
    end do
    do k = mm_KPol(i,j)+mm_KExp(i,j)+1, mm_KPol(i,j)+mm_KExp(i,j)+mm_KSpec(i,j)
      Val = Val &
        & + amm_BinDepFuncSmallN(k,i,j) &
        &   * Delta**amm_BinDepFuncSmallD(k,i,j) &
        &   * Tau**amm_BinDepFuncSmallT(k,i,j) &
        &   * exp( - amm_BinDepFuncEta (k,i,j) * ( Delta - amm_BinDepFuncEps  (k,i,j) )**2 &
        &          - amm_BinDepFuncBeta(k,i,j) * ( Delta - amm_BinDepFuncGamma(k,i,j) ) )
    end do


  end function BSDF

  !----------------------------------------------------------------------------
  ! delta derivative of binary specific departure function (BSDF)
  ! (modified version of Kunz and Wagner, 2012, Table B8)
  !----------------------------------------------------------------------------

  function DBSDFDDelta( Delta, Tau, i, j ) &
    & result( Val )

    real(dp), intent(in) :: Delta
    real(dp), intent(in) :: Tau
    integer , intent(in) :: i
    integer , intent(in) :: j
    real(dp)             :: Val


    ! Local variables
    !
    integer  :: k


    Val = 0.0d0

    do k = 1, mm_KPol(i,j)
      Val = Val &
        & + amm_BinDepFuncSmallN(k,i,j) &
        &   * amm_BinDepFuncSmallD(k,i,j) &
        &   * Delta**(amm_BinDepFuncSmallD(k,i,j)-1) &
        &   * Tau**amm_BinDepFuncSmallT(k,i,j)
    end do
    do k = mm_KPol(i,j)+1, mm_KPol(i,j)+mm_KExp(i,j)
      Val = Val &
        & + amm_BinDepFuncSmallN(k,i,j) &
        &   * Delta**amm_BinDepFuncSmallD(k,i,j) &
        &   * Tau**amm_BinDepFuncSmallT(k,i,j) &
        &   * exp( -Delta**amm_BinDepFuncSmallL(k,i,j) ) &
        &   * (   amm_BinDepFuncSmallD(k,i,j) / Delta &
        &       - amm_BinDepFuncSmallL(k,i,j) * Delta**(amm_BinDepFuncSmallL(k,i,j)-1.0d0) )
    end do
    do k = mm_KPol(i,j)+mm_KExp(i,j)+1, mm_KPol(i,j)+mm_KExp(i,j)+mm_KSpec(i,j)
      Val = Val &
        & + amm_BinDepFuncSmallN(k,i,j) &
        &   * Delta**amm_BinDepFuncSmallD(k,i,j) &
        &   * Tau**amm_BinDepFuncSmallT(k,i,j) &
        &   * exp( - amm_BinDepFuncEta (k,i,j) * ( Delta - amm_BinDepFuncEps  (k,i,j) )**2 &
        &          - amm_BinDepFuncBeta(k,i,j) * ( Delta - amm_BinDepFuncGamma(k,i,j) ) ) &
        &   * (   amm_BinDepFuncSmallD(k,i,j) / Delta &
        &       - 2.0d0 * amm_BinDepFuncEta (k,i,j) * ( Delta - amm_BinDepFuncEps  (k,i,j) ) &
        &       - amm_BinDepFuncBeta(k,i,j) )
    end do


  end function DBSDFDDelta

  !----------------------------------------------------------------------------
  ! delta second derivative of binary specific departure function (BSDF)
  ! (modified version of Kunz and Wagner, 2012, Table B8)
  !----------------------------------------------------------------------------

  function D2BSDFDDelta2( Delta, Tau, i, j ) &
    & result( Val )

    real(dp), intent(in) :: Delta
    real(dp), intent(in) :: Tau
    integer , intent(in) :: i
    integer , intent(in) :: j
    real(dp)             :: Val


    ! Local variables
    !
    integer  :: k


    Val = 0.0d0

    do k = 1, mm_KPol(i,j)
      Val = Val &
        & + amm_BinDepFuncSmallN(k,i,j) &
        &   * amm_BinDepFuncSmallD(k,i,j) &
        &   * ( amm_BinDepFuncSmallD(k,i,j) - 1 ) &
        &   * Delta**(amm_BinDepFuncSmallD(k,i,j)-2.0d0) &
        &   * Tau**amm_BinDepFuncSmallT(k,i,j)
    end do
    do k = mm_KPol(i,j)+1, mm_KPol(i,j)+mm_KExp(i,j)
      Val = Val &
        & + amm_BinDepFuncSmallN(k,i,j) &
        &   * Delta**amm_BinDepFuncSmallD(k,i,j) &
        &   * Tau**amm_BinDepFuncSmallT(k,i,j) &
        &   * exp( -Delta**amm_BinDepFuncSmallL(k,i,j) ) &
        &   * (   (   amm_BinDepFuncSmallD(k,i,j) / Delta &
        &           - amm_BinDepFuncSmallL(k,i,j) * Delta**(amm_BinDepFuncSmallL(k,i,j)-1.0d0) )**2 &
        &       - amm_BinDepFuncSmallD(k,i,j) / Delta**2 &
        &       - amm_BinDepFuncSmallL(k,i,j) * ( amm_BinDepFuncSmallL(k,i,j) - 1.0d0 ) &
        &         * Delta**(amm_BinDepFuncSmallL(k,i,j)-2.0d0) )
    end do
    do k = mm_KPol(i,j)+mm_KExp(i,j)+1, mm_KPol(i,j)+mm_KExp(i,j)+mm_KSpec(i,j)
      Val = Val &
        & + amm_BinDepFuncSmallN(k,i,j) &
        &   * Delta**amm_BinDepFuncSmallD(k,i,j) &
        &   * Tau**amm_BinDepFuncSmallT(k,i,j) &
        &   * exp( - amm_BinDepFuncEta (k,i,j) * ( Delta - amm_BinDepFuncEps  (k,i,j) )**2 &
        &          - amm_BinDepFuncBeta(k,i,j) * ( Delta - amm_BinDepFuncGamma(k,i,j) ) ) &
        &   * ( &
        &         (   amm_BinDepFuncSmallD(k,i,j) / Delta &
        &           - 2.0d0 * amm_BinDepFuncEta (k,i,j) * ( Delta - amm_BinDepFuncEps  (k,i,j) ) &
        &           - amm_BinDepFuncBeta(k,i,j) )**2 &
        &       - amm_BinDepFuncSmallD(k,i,j) / Delta**2 - 2.0d0 * amm_BinDepFuncEta (k,i,j) &
        &     )
    end do


  end function D2BSDFDDelta2

  !----------------------------------------------------------------------------
  ! tau derivative of binary specific departure function (BSDF)
  ! (modified version of Kunz and Wagner, 2012, Table B8)
  !----------------------------------------------------------------------------

  function DBSDFDTau( Delta, Tau, i, j ) &
    & result( Val )

    real(dp), intent(in) :: Delta
    real(dp), intent(in) :: Tau
    integer , intent(in) :: i
    integer , intent(in) :: j
    real(dp)             :: Val


    ! Local variables
    !
    integer  :: k


    Val = 0.0d0

    do k = 1, mm_KPol(i,j)
      Val = Val &
        & + amm_BinDepFuncSmallN(k,i,j) &
        &   * amm_BinDepFuncSmallT(k,i,j) &
        &   * Delta**amm_BinDepFuncSmallD(k,i,j) &
        &   * Tau**(amm_BinDepFuncSmallT(k,i,j)-1.0d0)
    end do
    do k = mm_KPol(i,j)+1, mm_KPol(i,j)+mm_KExp(i,j)
      Val = Val &
        & + amm_BinDepFuncSmallN(k,i,j) &
        &   * amm_BinDepFuncSmallT(k,i,j) &
        &   * Delta**amm_BinDepFuncSmallD(k,i,j) &
        &   * Tau**(amm_BinDepFuncSmallT(k,i,j)-1.0d0) &
        &   * exp( -Delta**amm_BinDepFuncSmallL(k,i,j) )
    end do
    do k = mm_KPol(i,j)+mm_KExp(i,j)+1, mm_KPol(i,j)+mm_KExp(i,j)+mm_KSpec(i,j)
      Val = Val &
        & + amm_BinDepFuncSmallN(k,i,j) &
        &   * amm_BinDepFuncSmallT(k,i,j) &
        &   * Delta**amm_BinDepFuncSmallD(k,i,j) &
        &   * Tau**(amm_BinDepFuncSmallT(k,i,j)-1.0d0) &
        &   * exp( - amm_BinDepFuncEta (k,i,j) * ( Delta - amm_BinDepFuncEps  (k,i,j) )**2 &
        &          - amm_BinDepFuncBeta(k,i,j) * ( Delta - amm_BinDepFuncGamma(k,i,j) ) )
    end do


  end function DBSDFDTau

  !----------------------------------------------------------------------------
  ! tau second derivative of binary specific departure function (BSDF)
  ! (modified version of Kunz and Wagner, 2012, Table B8)
  !----------------------------------------------------------------------------

  function D2BSDFDTau2( Delta, Tau, i, j ) &
    & result( Val )

    real(dp), intent(in) :: Delta
    real(dp), intent(in) :: Tau
    integer , intent(in) :: i
    integer , intent(in) :: j
    real(dp)             :: Val


    ! Local variables
    !
    integer  :: k


    Val = 0.0d0

    do k = 1, mm_KPol(i,j)
      Val = Val &
        & + amm_BinDepFuncSmallN(k,i,j) &
        &   * amm_BinDepFuncSmallT(k,i,j) &
        &   * ( amm_BinDepFuncSmallT(k,i,j) - 1.0d0 ) &
        &   * Delta**amm_BinDepFuncSmallD(k,i,j) &
        &   * Tau**(amm_BinDepFuncSmallT(k,i,j)-2.0d0)
    end do
    do k = mm_KPol(i,j)+1, mm_KPol(i,j)+mm_KExp(i,j)
      Val = Val &
        & + amm_BinDepFuncSmallN(k,i,j) &
        &   * amm_BinDepFuncSmallT(k,i,j) &
        &   * ( amm_BinDepFuncSmallT(k,i,j) - 1.0d0 ) &
        &   * Delta**amm_BinDepFuncSmallD(k,i,j) &
        &   * Tau**(amm_BinDepFuncSmallT(k,i,j)-2.0d0) &
        &   * exp( -Delta**amm_BinDepFuncSmallL(k,i,j) )
    end do
    do k = mm_KPol(i,j)+mm_KExp(i,j)+1, mm_KPol(i,j)+mm_KExp(i,j)+mm_KSpec(i,j)
      Val = Val &
        & + amm_BinDepFuncSmallN(k,i,j) &
        &   * amm_BinDepFuncSmallT(k,i,j) &
        &   * ( amm_BinDepFuncSmallT(k,i,j) - 1.0d0 ) &
        &   * Delta**amm_BinDepFuncSmallD(k,i,j) &
        &   * Tau**(amm_BinDepFuncSmallT(k,i,j)-2.0d0) &
        &   * exp( - amm_BinDepFuncEta (k,i,j) * ( Delta - amm_BinDepFuncEps  (k,i,j) )**2 &
        &          - amm_BinDepFuncBeta(k,i,j) * ( Delta - amm_BinDepFuncGamma(k,i,j) ) )
    end do


  end function D2BSDFDTau2

  !----------------------------------------------------------------------------
  ! delta and tau derivative of binary specific departure function (BSDF)
  ! (modified version of Kunz and Wagner, 2012, Table B8)
  !----------------------------------------------------------------------------

  function D2BSDFDDeltaDTau( Delta, Tau, i, j ) &
    & result( Val )

    real(dp), intent(in) :: Delta
    real(dp), intent(in) :: Tau
    integer , intent(in) :: i
    integer , intent(in) :: j
    real(dp)             :: Val


    ! Local variables
    !
    integer  :: k


    Val = 0.0d0

    do k = 1, mm_KPol(i,j)
      Val = Val &
        & + amm_BinDepFuncSmallN(k,i,j) &
        &   * amm_BinDepFuncSmallD(k,i,j) &
        &   * amm_BinDepFuncSmallT(k,i,j) &
        &   * Delta**(amm_BinDepFuncSmallD(k,i,j)-1) &
        &   * Tau**(amm_BinDepFuncSmallT(k,i,j)-1.0d0)
    end do
    do k = mm_KPol(i,j)+1, mm_KPol(i,j)+mm_KExp(i,j)
      Val = Val &
        & + amm_BinDepFuncSmallN(k,i,j) &
        &   * amm_BinDepFuncSmallT(k,i,j) &
        &   * Delta**amm_BinDepFuncSmallD(k,i,j) &
        &   * Tau**(amm_BinDepFuncSmallT(k,i,j)-1.0d0) &
        &   * exp( -Delta**amm_BinDepFuncSmallL(k,i,j) ) &
        &   * (   amm_BinDepFuncSmallD(k,i,j) / Delta &
        &       - amm_BinDepFuncSmallL(k,i,j) * Delta**(amm_BinDepFuncSmallL(k,i,j)-1.0d0) )
    end do
    do k = mm_KPol(i,j)+mm_KExp(i,j)+1, mm_KPol(i,j)+mm_KExp(i,j)+mm_KSpec(i,j)
      Val = Val &
        & + amm_BinDepFuncSmallN(k,i,j) &
        &   * amm_BinDepFuncSmallT(k,i,j) &
        &   * Delta**amm_BinDepFuncSmallD(k,i,j) &
        &   * Tau**(amm_BinDepFuncSmallT(k,i,j)-1.0d0) &
        &   * exp( - amm_BinDepFuncEta (k,i,j) * ( Delta - amm_BinDepFuncEps  (k,i,j) )**2 &
        &          - amm_BinDepFuncBeta(k,i,j) * ( Delta - amm_BinDepFuncGamma(k,i,j) ) ) &
        &   * (   amm_BinDepFuncSmallD(k,i,j) / Delta &
        &       - 2.0d0 * amm_BinDepFuncEta (k,i,j) * ( Delta - amm_BinDepFuncEps  (k,i,j) ) &
        &       - amm_BinDepFuncBeta(k,i,j) )
    end do


  end function D2BSDFDDeltaDTau

  !----------------------------------------------------------------------------

end module thermo_prop_mix_gs2016_co2n2
