! EOS is that of van der Waals equation.
! Internal energy of ideal gas component is obtained from the EOS-CG.
module thermo_prop_vdWEOS

  use vtype_module

  implicit none

  private

  character(stdstr), parameter :: ModuleName = 'thermo_prop_vdWEOS'

  real(dp), save :: ConvThreshold = 1.0d-11

  data ConvThreshold /1.0d-11/

  real(dp), parameter :: UnivGasConst = 8.314510d0  ! J mole-1 K-1
  real(dp), save      :: MolWeight
  real(dp), save      :: GasConst


  real(dp), save :: vdWCoefA0
  real(dp), save :: vdWCoefB0

  real(dp), save :: vdWCoefA
  real(dp), save :: vdWCoefB

  ! critical point
  real(dp), save :: DensC
  real(dp), save :: TempC
  real(dp), save :: PressC


  ! DelEntrMin is defined to avoid aborting the program.
  ! See memo in IR_FindDelEntropyForTempEq.
  !
  real(dp), parameter :: DelEntrMin = 1.0d-5


!!$  integer , parameter :: NCoef = 3
!!$  real(dp), save      :: a_CvCoef(NCoef)
!!$  real(dp), parameter :: Temp1 = 450.0d0
!!$  real(dp), parameter :: Temp2 = 300.0d0
!!$
!!$  ! Cv =   a_CvCoef(1) * ((T-Temp1)/Temp2)**(1-1)
!!$  !      + a_CvCoef(2) * ((T-Temp1)/Temp2)**(2-1)
!!$  !      + a_CvCoef(3) * ((T-Temp1)/Temp2)**(3-1)
!!$  data a_CvCoef / 791.3d0, 235.9d0, -65.57d0 /


  ! CO2
  !   a = 27 R^2 Tc^2 / ( 64 Pc )   (Pa m6 mol-2)
  !   b = R Tc / ( 8 Pc )           (m3 mol-1)
  !
  !   From Handbook of Chemistry: Pure Chemistry, 5th ed.
  !     for CO2
  !       Tc   =  304.21     K
  !       Pc   =    7.3825e6 Pa
  !       Rhoc = 4661        kg m-3
  !       a    = 3.66e-1
  !       b    = 4.28e-5
  !     for N2
  !       Tc   =  126.2     K
  !       Pc   =    3.4e6   Pa
  !       Rhoc =  314       kg m-3
  !       a    = 1.37e-1
  !       b    = 3.86e-5
  !
  integer          , parameter :: NSubst = 3
  character(stdstr), save      :: m_SubstName     (NSubst)
  real(dp)         , save      :: m_SubstVMR      (NSubst)
  real(dp)         , save      :: m_SubstMolWeight(NSubst)
  real(dp)         , save      :: m_SubstCoefA0   (NSubst)
  real(dp)         , save      :: m_SubstCoefB0   (NSubst)

  data m_SubstName      / "CO2", "H2O", "N2" /
  data m_SubstVMR       / 0.965d0   , -1.0d100,   0.035d0    /
  data m_SubstMolWeight / 44.0098d-3, -1.0d100,  28.01348d-3 /
  data m_SubstCoefA0    / 3.66d-1   , -1.0d100,   1.37d-1    /
  data m_SubstCoefB0    / 4.28d-5   , -1.0d100,   3.86d-5    /

  public :: ThermoPropInqFlagFromDensTemp
  public :: ThermoPropInqFlagFromDensPress

  public :: ThermoPropInit
  public :: ThermoPropInqMolGasConst
  public :: ThermoPropInqMolWeight
  public :: ThermoPropInqTempCrit
  public :: ThermoPropInqPressCrit
  public :: ThermoPropInqDensCrit
  public :: ThermoPropCompFact
  public :: ThermoPropDens
  public :: ThermoPropTemp
  public :: ThermoPropTempFromPressEnthalpy
  public :: ThermoPropEnthalpy
  public :: ThermoPropCp
  public :: ThermoPropDPressDDensConstTemp
  public :: ThermoPropDPressDTempConstDens
  public :: ThermoPropThermoExpCoef
  public :: ThermoPropAdLapRateDTDP
  public :: ThermoPropDensTempFromPressEnthalpy


  !----------------------------------------------------------------------------

contains

  !----------------------------------------------------------------------------
  ! set parameters
  !----------------------------------------------------------------------------

  subroutine ThermoPropInit(     &
    & ConvThresholdSet           &
    & )

    use thermo_prop_mix_gs2016_co2n2, only : &
      & ThermoPropInitGS2016 => ThermoPropInit


    real(dp), intent(in), optional :: ConvThresholdSet

    real(dp) :: SumVMR
    integer  :: i
    integer  :: j


    if ( present( ConvThresholdSet ) ) then
      ConvThreshold = ConvThresholdSet
    end if

    ! check volume mixing ratio
    SumVMR = 0.0d0
    do i = 1, NSubst
      if ( m_SubstVMR(i) > 0.0d0 ) then
        SumVMR = SumVMR + m_SubstVMR(i)
      end if
    end do
    if ( SumVMR /= 1.0d0 ) then
      write( 6, * ) 'Error : Sum of volume mixing ratio is not 1, ', SumVMR
      stop
    end if

    ! mean molecular weight
    MolWeight = 0.0d0
    do i = 1, NSubst
      if ( m_SubstVMR(i) > 0.0d0 ) then
        MolWeight = MolWeight + m_SubstVMR(i) * m_SubstMolWeight(i)
      end if
    end do

    ! gas constant
    GasConst = UnivGasConst / MolWeight ! J kg-1 K-1

    ! Initialization of module used in this module.

    ! Internal enery for ideal gas is used in this routine.
    call ThermoPropInitGS2016( &
      & NSubst, m_SubstName, m_SubstVMR, &
      & FlagIdealGasSet  = .true.,       &
      & ConvThresholdSet = ConvThreshold &
      & )

    ! Coefficients for van der Waals EOS
    vdWCoefA0 = 0.0d0
    vdWCoefB0 = 0.0d0
    do i = 1, NSubst
      if ( m_SubstVMR(i) > 0.0d0 ) then
        do j = 1, NSubst
          if ( m_SubstVMR(j) > 0.0d0 ) then
            vdWCoefA0 = vdWCoefA0 &
              & + m_SubstVMR(i) * m_SubstVMR(j) &
              &   * ( 1.0d0 - 0.0d0 ) & 
              &   * sqrt( m_SubstCoefA0(i) * m_SubstCoefA0(j) )
            vdWCoefB0 = vdWCoefB0 &
              & + m_SubstVMR(i) * m_SubstVMR(j) &
              &   * ( m_SubstCoefB0(i) + m_SubstCoefB0(j) ) / 2.0d0
          end if
        end do
      end if
    end do

    !
    ! The values below are those for least squared fit.
    !
!!$    vdWCoefA0 = AAAAA
!!$    vdWCoefB0 = BBBBB
    !
    ! The values below are calculated by script/lsf_ab.rb with above values.
    ! 10: vdWA=0.3735237597918917, vdWB=5.974492586635085e-05, error=1.7112071512290045e-05 1
    !
!!$    vdWCoefA0 = 3.735d-1
!!$    vdWCoefB0 = 5.974d-5

!!$    vdWCoefA0 = vdWCoefA0 * 0.9d0
!!$    vdWCoefB0 = vdWCoefB0


    ! coefficients per unit mass
    vdWCoefA  = vdWCoefA0 / MolWeight**2
    vdWCoefB  = vdWCoefB0 / MolWeight

    ! critical point
    DensC  = 1.0d0/(3.0d0*vdWCoefB)                    ! rho_c (kg m-3)
    TempC  = 8.0d0*vdWCoefA/(27.0d0*vdWCoefB*GasConst) ! T_c (K)
    PressC = vdWCoefA/(27.0d0*vdWCoefB**2)             ! p_c (Pa)


  end subroutine ThermoPropInit

  !----------------------------------------------------------------------------
  ! inquire flag for calculation
  !----------------------------------------------------------------------------
  ! If FlagOnlyPhaseEq is .true., .true. is returned when the density is
  ! greater than phase equilibrium.
  !----------------------------------------------------------------------------

  function ThermoPropInqFlagFromDensTemp( Dens, Temp, FlagExcludeOnlyPhaseEq ) &
    & result( Val )

    real(dp), intent(in)           :: Dens
    real(dp), intent(in)           :: Temp
    logical , intent(in), optional :: FlagExcludeOnlyPhaseEq

    logical                        :: Val


    ! Local variables
    !
    logical  :: LV_FlagExcludeOnlyPhaseEq
    real(dp) :: Delta
    real(dp) :: Tau
    real(dp) :: DelEntr
    real(dp) :: DeltaLiq
    real(dp) :: DeltaGas


    if ( present( FlagExcludeOnlyPhaseEq ) ) then
      LV_FlagExcludeOnlyPhaseEq = FlagExcludeOnlyPhaseEq
    else
      LV_FlagExcludeOnlyPhaseEq = .true.
    end if


    Delta = Dens / DensC
    Tau   = Temp / TempC

    if ( Tau >= 1.0d0 ) then

      Val = .true.

    else

      DelEntr = IR_FindDelEntropyForTempEq( Temp )
      call IR_vdWEOSRedDensLimAtPhaseEq( DelEntr/2.0d0, DeltaLiq, DeltaGas )

      if ( Delta <= DeltaGas ) then

        ! Region of density lower than that in phase equilibrium

        Val = .true.

      else if ( Delta <= DeltaLiq ) then

        ! Region of density in phase equilibrium

        Val = .false.

      else

        ! Region of density higher than that in phase equilibrium

        if ( LV_FlagExcludeOnlyPhaseEq ) then
          Val = .true.
        else
          Val = .false.
        end if

      end if

    end if


  end function ThermoPropInqFlagFromDensTemp

  !----------------------------------------------------------------------------
  ! temperature from pressure and enthalpy
  !
  ! This routine calculate temperature by the use of pressure and enthalpy.
  ! But, this routine requires a very small threshold for convergence,
  ! ConvThreshold ~ 1e-9 or 1e-10 or so.
  ! Instead, a routine using ThermoPropDensTempFromPressEnthalpy seems to
  ! calculate better temperatures with larger threshold, ConvThreshold ~ 1e-5.
  !----------------------------------------------------------------------------

  function NG_ThermoPropTempFromPressEnthalpy( Press, Enthalpy ) & ! (in)
    & result( Temp )

    real(dp), intent(in ) :: Press
    real(dp), intent(in ) :: Enthalpy
    real(dp)              :: Temp


    ! Local variables
    !
!!$    real(dp) :: TempMin = 60.0d0
    real(dp) :: Dens
    real(dp) :: TempCur
!!$    real(dp) :: TempNext
    real(dp) :: Temp1
    real(dp) :: Temp2
    real(dp) :: Enthalpy1
    real(dp) :: Enthalpy2
    real(dp) :: EnthalpyCur
!!$    real(dp) :: DHDTCP
!!$    real(dp) :: DelTemp
!!$    integer, parameter  :: NLoop = 100
!!$    integer             :: iLoop
!!$    real(dp)            :: a_HistDens( 0:NLoop )
!!$    real(dp)            :: a_HistTemp( 0:NLoop )


    ! Initialization
    !   find temperature with a given pressure

    ! Bisectional method
    Temp1     = 50.0d0
    Dens      = ThermoPropDens( Press, Temp1 )
    Enthalpy1 = ThermoPropEnthalpy( Dens, Temp1 )
    if ( Enthalpy < Enthalpy1 ) then
      write( 6, * ) 'ERROR: In ', trim( ModuleName )
      write( 6, * ) 'ERROR: In ThermoPropTempFromPressEnthalpy'
      write( 6, * ) 'ERROR:   Temperature seems to be too low.'
      stop
    end if
    search_init_temperature : do
      if ( Temp1 < 100.0d0 ) then
        Temp2     = Temp1 + 10.0d0
      else
        Temp2     = Temp1 + 100.0d0
      end if
      Dens      = ThermoPropDens( Press, Temp2 )
      Enthalpy2 = ThermoPropEnthalpy( Dens, Temp2 )
      if ( Enthalpy < Enthalpy2 ) then
        exit
      end if
      Temp1 = Temp2
    end do search_init_temperature
    bisectional_method : do
      TempCur     = ( Temp1 + Temp2 ) / 2.0d0
      Dens        = ThermoPropDens( Press, TempCur )
      EnthalpyCur = ThermoPropEnthalpy( Dens, TempCur )
      if ( Enthalpy < EnthalpyCur ) then
        Temp2 = TempCur
      else
        Temp1 = TempCur
      end if
      if ( abs( ( Temp2 - Temp1 ) / Temp1 ) < ConvThreshold ) then
        exit bisectional_method
      end if
    end do bisectional_method
    Temp = ( Temp1 + Temp2 ) / 2.0d0


    ! Newton's method
!!$    TempCur = TempMin
!!$    a_HistTemp(0) = TempCur
!!$    search_temperature : do iLoop = 1, NLoop
!!$      Dens        = ThermoPropDens( Press, TempCur )
!!$      EnthalpyCur = ThermoPropEnthalpy( Dens, TempCur )
!!$      DHDTCP      = ThermoPropCp( Dens, TempCur )
!!$      if ( DHDTCP <= 0.0d0 ) then
!!$        write( 6, * ) 'ERROR: In ', trim( ModuleName )
!!$        write( 6, * ) 'ERROR: In ThermoPropTempFromPressEnthalpy'
!!$        write( 6, * ) 'ERROR:   (DHDT)_p is negative, ', DHDTCP
!!$        write( 6, * ) 'ERROR:   at rho = ', Dens, ', T = ', TempCur
!!$        stop
!!$      end if
!!$      DelTemp     = ( Enthalpy - EnthalpyCur ) / DHDTCP
!!$      TempNext    = TempCur + DelTemp
!!$      TempNext    = max( TempNext, TempMin )
!!$      a_HistTemp( iLoop ) = TempNext
!!$      if ( abs( ( TempNext - TempCur ) / TempCur ) < ConvThreshold ) then
!!$        exit search_temperature
!!$      end if
!!$      TempCur = TempNext
!!$    end do search_temperature
!!$    if ( iLoop >= NLoop ) then
!!$      write( 6, * ) 'ERROR: In ', trim( ModuleName )
!!$      write( 6, * ) 'ERROR: In ThermoPropTempFromPressEnthalpy'
!!$      write( 6, * ) 'ERROR:   Loop for temperature does not converge'
!!$      write( 6, * ) 'ERROR:   ', iLoop
!!$      write( 6, * ) 'ERROR:   ', a_HistTemp
!!$      stop
!!$    end if
!!$    Temp = TempNext
!!$
!!$    if ( Temp == TempMin ) then
!!$      write( 6, * ) 'ERROR: In ', trim( ModuleName )
!!$      write( 6, * ) 'ERROR: In ThermoPropTempFromPressEnthalpy'
!!$      write( 6, * ) 'ERROR:   Temperature seems to be calculated, properly.'
!!$      write( 6, * ) 'ERROR:   ', Temp, TempMin
!!$      stop
!!$    end if


  end function NG_ThermoPropTempFromPressEnthalpy

  !----------------------------------------------------------------------------
  ! temperature from pressure and enthalpy
  !----------------------------------------------------------------------------

  function ThermoPropTempFromPressEnthalpy( Press, Enthalpy ) & ! (in)
    & result( Temp )

    real(dp), intent(in ) :: Press
    real(dp), intent(in ) :: Enthalpy
    real(dp)              :: Temp


    ! Local variables
    !
    real(dp) :: Dens


    call ThermoPropDensTempFromPressEnthalpy( &
      & Press, Enthalpy, & ! (in)
      & Dens, Temp       & ! (out)
      & )


  end function ThermoPropTempFromPressEnthalpy

  !----------------------------------------------------------------------------
  ! inquire flag for calculation
  !----------------------------------------------------------------------------
  ! If FlagOnlyPhaseEq is .true., .true. is returned when the density is
  ! greater than phase equilibrium.
  !----------------------------------------------------------------------------

  function ThermoPropInqFlagFromDensPress( Dens, Press, FlagExcludeOnlyPhaseEq ) &
    & result( Val )

    real(dp), intent(in)           :: Dens
    real(dp), intent(in)           :: Press
    logical , intent(in), optional :: FlagExcludeOnlyPhaseEq

    logical                        :: Val


    ! Local variables
    !
    logical  :: LV_FlagExcludeOnlyPhaseEq
    real(dp) :: Delta
    real(dp) :: Pi
    real(dp) :: DelEntr
    real(dp) :: DeltaLiq
    real(dp) :: DeltaGas


    if ( present( FlagExcludeOnlyPhaseEq ) ) then
      LV_FlagExcludeOnlyPhaseEq = FlagExcludeOnlyPhaseEq
    else
      LV_FlagExcludeOnlyPhaseEq = .true.
    end if


    Delta = Dens  / DensC
    Pi    = Press / PressC

    if ( Pi >= 1.0d0 ) then

      Val = .true.

    else

      DelEntr = IR_FindDelEntropyForPressEq( Press )
      call IR_vdWEOSRedDensLimAtPhaseEq( DelEntr/2.0d0, DeltaLiq, DeltaGas )

      if ( Delta <= DeltaGas ) then

        ! Region of density lower than that in phase equilibrium

        Val = .true.

      else if ( Delta <= DeltaLiq ) then

        ! Region of density in phase equilibrium

        Val = .false.

      else

        ! Region of density higher than that in phase equilibrium

        if ( LV_FlagExcludeOnlyPhaseEq ) then
          Val = .true.
        else
          Val = .false.
        end if

      end if

    end if


  end function ThermoPropInqFlagFromDensPress

  !----------------------------------------------------------------------------
  ! inquire molar gas constant
  !----------------------------------------------------------------------------

  function ThermoPropInqMolGasConst() &
    & result( Val )

    real(dp) :: Val

    Val = UnivGasConst

  end function ThermoPropInqMolGasConst

  !----------------------------------------------------------------------------
  ! inquire molecular weight
  !----------------------------------------------------------------------------

  function ThermoPropInqMolWeight() &
    & result( Val )

    real(dp) :: Val

    Val = MolWeight

  end function ThermoPropInqMolWeight

  !----------------------------------------------------------------------------
  ! inquire temperature of critical point
  !----------------------------------------------------------------------------

  function ThermoPropInqTempCrit() &
    & result( Val )

    real(dp) :: Val

    Val = TempC

  end function ThermoPropInqTempCrit

  !----------------------------------------------------------------------------
  ! inquire pressure of critical point
  !----------------------------------------------------------------------------

  function ThermoPropInqPressCrit() &
    & result( Val )

    real(dp) :: Val

    Val = PressC

  end function ThermoPropInqPressCrit

  !----------------------------------------------------------------------------
  ! inquire temperature of critical point
  !----------------------------------------------------------------------------

  function ThermoPropInqDensCrit() &
    & result( Val )

    real(dp) :: Val

    Val = DensC

  end function ThermoPropInqDensCrit

  !----------------------------------------------------------------------------
  ! Internal routine
  ! Calculate reduced pressure of van der Waals EOS
  !----------------------------------------------------------------------------

  function IR_vdWEOSRedPressOrg( Delta, Tau ) &
    result( Val )

    real(dp), intent(in) :: Delta
    real(dp), intent(in) :: Tau

    real(dp) :: Val


    Val = 8.0d0 * Delta * Tau / ( 3.0d0 - Delta ) - 3.0d0 * Delta**2


  end function IR_VdWEOSRedPressOrg

  !----------------------------------------------------------------------------
  ! Internal routine
  ! Calculate reduced temperature of van der Waals EOS
  !----------------------------------------------------------------------------

  function IR_vdWEOSRedTempOrg( Delta, Pi ) &
    result( Val )

    real(dp), intent(in) :: Delta
    real(dp), intent(in) :: Pi

    real(dp) :: Val


    Val = ( 3.0d0 - Delta ) * ( Pi + 3.0d0 * Delta**2 ) / ( 8.0d0 * Delta )


  end function IR_VdWEOSRedTempOrg

  !----------------------------------------------------------------------------
  ! Internal routine
  ! function f(x)
  ! It should be noticed that x = y/2, where y is defined at
  ! https://farside.ph.utexas.edu/teaching/sm1/Thermalhtml/node124.html#ex9.7
  ! and x = \Delta x / 2, where \Delta x is dfined by Lekner (1982).
  !----------------------------------------------------------------------------

  function IR_FuncF( X ) &
    & result( Val )

    real(dp), intent(in) :: X

    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: C
    real(dp) :: S


    C = cosh(X)
    S = sinh(X)
    Val = ( X * C - S ) / ( C * S - X )


  end function IR_FuncF

  !----------------------------------------------------------------------------
  ! Internal routine
  ! function c(x)
  ! It should be noticed that x = y/2, where y is defined at
  ! https://farside.ph.utexas.edu/teaching/sm1/Thermalhtml/node124.html#ex9.7
  ! and x = \Delta x / 2, where \Delta x is dfined by Lekner (1982).
  !----------------------------------------------------------------------------

  function IR_FuncC( X ) &
    & result( Val )

    real(dp), intent(in) :: X

    real(dp)             :: Val


    Val = cosh(X)


  end function IR_FuncC

  !----------------------------------------------------------------------------
  ! Internal routine
  ! function g(x)
  ! It should be noticed that x = y/2, where y is defined at
  ! https://farside.ph.utexas.edu/teaching/sm1/Thermalhtml/node124.html#ex9.7
  ! and x = \Delta x / 2, where \Delta x is dfined by Lekner (1982).
  !----------------------------------------------------------------------------

  function IR_FuncG( X ) &
    & result( Val )

    real(dp), intent(in) :: X

    real(dp)             :: Val

    ! Local variables
    !
    real(dp) :: C
    real(dp) :: F


    C = IR_FuncC(X)
    F = IR_FuncF(X)
    Val = 1.0d0 + 2.0d0 * C * F + F**2


  end function IR_FuncG

  !----------------------------------------------------------------------------
  ! Internal routine
  ! function dfdx(x)
  ! It should be noticed that x = y/2, where y is defined at
  ! https://farside.ph.utexas.edu/teaching/sm1/Thermalhtml/node124.html#ex9.7
  ! and x = \Delta x / 2, where \Delta x is dfined by Lekner (1982).
  !----------------------------------------------------------------------------

  function IR_FuncDFDX( X ) &
    & result( Val )

    real(dp), intent(in) :: X

    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: C
    real(dp) :: S


    C = cosh(X)
    S = sinh(X)

    Val =   ( &
      &         ( C + X * S - C ) * ( C * S - X ) &
      &       - ( X * C - S ) * ( S**2 + C**2 - 1.0d0 ) &
      &     ) &
      &   / ( C * S - X )**2


  end function IR_FuncDFDX

  !----------------------------------------------------------------------------
  ! Internal routine
  ! function dcdx(x)
  ! It should be noticed that x = y/2, where y is defined at
  ! https://farside.ph.utexas.edu/teaching/sm1/Thermalhtml/node124.html#ex9.7
  ! and x = \Delta x / 2, where \Delta x is dfined by Lekner (1982).
  !----------------------------------------------------------------------------

  function IR_FuncDCDX( X ) &
    & result( Val )

    real(dp), intent(in) :: X

    real(dp)             :: Val


    Val = sinh(X)


  end function IR_FuncDCDX

  !----------------------------------------------------------------------------
  ! Internal routine
  ! function dgdx(x)
  ! It should be noticed that x = y/2, where y is defined at
  ! https://farside.ph.utexas.edu/teaching/sm1/Thermalhtml/node124.html#ex9.7
  ! and x = \Delta x / 2, where \Delta x is dfined by Lekner (1982).
  !----------------------------------------------------------------------------

  function IR_FuncDGDX( X ) &
    & result( Val )

    real(dp), intent(in) :: X

    real(dp)             :: Val

    ! Local variables
    !
    real(dp) :: C
    real(dp) :: F
    real(dp) :: DCDX
    real(dp) :: DFDX


    C    = IR_FuncC(X)
    F    = IR_FuncF(X)
    DCDX = IR_FuncDCDX(X)
    DFDX = IR_FuncDFDX(X)
    Val = 2.0d0 * DCDX * F + 2.0d0 * C * DFDX + 2.0d0 * F * DFDX


  end function IR_FuncDGDX

  !----------------------------------------------------------------------------
  ! Internal routine
  ! function PressPhaseEq(x)
  ! It should be noticed that x = y/2, where y is defined at
  ! https://farside.ph.utexas.edu/teaching/sm1/Thermalhtml/node124.html#ex9.7
  ! and x = \Delta x / 2, where \Delta x is dfined by Lekner (1982).
  !----------------------------------------------------------------------------

  function IR_vdWEOSRedPressAtPhaseEq( X ) &
    & result( Val )

    real(dp), intent(in) :: X

    real(dp)             :: Val

    ! Local variables
    !
    real(dp) :: F
    real(dp) :: G


    F = IR_FuncF(X)
    G = IR_FuncG(X)
    Val = 27.0d0 * F**2 * ( 1.0d0 - F**2 ) / G**2


  end function IR_VdWEOSRedPressAtPhaseEq

  !----------------------------------------------------------------------------
  ! Internal routine
  ! function TempPhaseEq(x)
  ! It should be noticed that x = y/2, where y is defined at
  ! https://farside.ph.utexas.edu/teaching/sm1/Thermalhtml/node124.html#ex9.7
  ! and x = \Delta x / 2, where \Delta x is dfined by Lekner (1982).
  !----------------------------------------------------------------------------

  function IR_vdWEOSRedTempAtPhaseEq( X ) &
    & result( Val )

    real(dp), intent(in) :: X

    real(dp)             :: Val

    ! Local variables
    !
    real(dp) :: C
    real(dp) :: F
    real(dp) :: G


    C = IR_FuncC(X)
    F = IR_FuncF(X)
    G = IR_FuncG(X)
    Val = 27.0d0 * F * ( F + C ) / ( 4.0d0 * G**2 )


  end function IR_VdWEOSRedTempAtPhaseEq

  !----------------------------------------------------------------------------
  ! Internal routine
  ! function vdWEOSRedDensLimAtPhaseEq(x)
  ! It should be noticed that x = y/2, where y is defined at
  ! https://farside.ph.utexas.edu/teaching/sm1/Thermalhtml/node124.html#ex9.7
  ! and x = \Delta x / 2, where \Delta x is dfined by Lekner (1982).
  !----------------------------------------------------------------------------

  subroutine IR_vdWEOSRedDensLimAtPhaseEq( X, RedDensLiq, RedDensGas )

    real(dp), intent(in ) :: X
    real(dp), intent(out) :: RedDensLiq
    real(dp), intent(out) :: RedDensGas


    ! Local variables
    !
    real(dp) :: F
    real(dp) :: Xplus
    real(dp) :: Xminus
    real(dp) :: RedVLiq
    real(dp) :: RedVGas


    F       = IR_FuncF(X)
    Xplus   = exp(  X ) * F
    Xminus  = exp( -X ) * F
    RedVLiq = ( 1.0d0 / Xplus  + 1.0d0 ) / 3.0d0
    RedVGas = ( 1.0d0 / Xminus + 1.0d0 ) / 3.0d0

    RedDensLiq = 1.0d0 / RedVLiq
    RedDensGas = 1.0d0 / RedVGas


  end subroutine IR_VdWEOSRedDensLimAtPhaseEq

  !----------------------------------------------------------------------------
  ! Internal routine
  ! Find entropy difference for a specified equilibrium temperature
  !----------------------------------------------------------------------------

  function IR_FindDelEntropyForTempEq( TempEq ) &
    & result( Val )

    real(dp), intent(in) :: TempEq

    real(dp)             :: Val

    ! Local variables
    !
    real(dp) :: TauEq
    real(dp) :: DelEntr1
    real(dp) :: DelEntr2
    real(dp) :: DelEntr
    real(dp) :: Tau1
    real(dp) :: Tau2
    real(dp) :: Tau

    integer , parameter :: NLoop = 100
    integer             :: iLoop
    real(dp)            :: a_Hist( 0:NLoop )


    TauEq = TempEq / TempC

    if ( TauEq > 1.0d0 ) then

      write( 6, * ) 'ERROR: In ', trim( ModuleName )
      write( 6, * ) 'ERROR: In IR_FindDelEntropyForTempEq'
      write( 6, * ) '       Specified temperature is higher than the critical temperature.'
      write( 6, * ) '       ', TempEq, TempC
      stop

    else

      ! Initialization
      !   Empirically, 1.0d-5 seems to be minimum of DelEntr1.
      !   The smaller the DelEntr is, the higher the temperature
      !   at phase equilibrium (TauEq) is.
      !   But, this is not the case for DelEntr1 less than 1.0d-5.
      !   So, DelEntrMin is defined as 1.0d-5. 
      DelEntr1 = DelEntrMin
      DelEntr2 = 7.0d2

      DelEntr = DelEntr1
      Tau     = IR_vdWEOSRedTempAtPhaseEq( DelEntr/2.0d0 )
      Tau1    = Tau
      !
      DelEntr = DelEntr2
      Tau     = IR_vdWEOSRedTempAtPhaseEq( DelEntr/2.0d0 )
      Tau2    = Tau

      if ( Tau1 < TauEq ) then
        ! If TauEq is larger than Tau1 which can be achieved in this program,
        ! DelEntr is set as DelEntrMin.

        DelEntr = DelEntrMin

      else

        ! The range is checked, again.
        if ( ( TauEq < Tau2 ) .or. ( Tau1 < TauEq ) ) then
          write( 6, * ) 'ERROR: In ', trim( ModuleName )
          write( 6, * ) 'ERROR: In IR_FindDelEntropyForTempEq'
          write( 6, * ) '       Initial range does not include specified temperature.'
          write( 6, * ) '       In redueced values: ', TauEq, Tau1, Tau2
          stop
        end if

        iLoop         = 0
        a_Hist(iLoop) = Tau1
        search_entropy : do iLoop = 1, NLoop
          DelEntr = ( DelEntr1 + DelEntr2 ) / 2.0d0
          Tau  = IR_vdWEOSRedTempAtPhaseEq( DelEntr/2.0d0 )

          if ( (Tau-TauEq) * (Tau1-TauEq) <= 0.0d0 ) then
            ! Target temperature is located between DelEntr1 and DelEntr.
            DelEntr2 = DelEntr
            Tau2     = Tau
          else
            ! Target temperature is located between DelEntr and DelEntr2.
            DelEntr1 = DelEntr
            Tau1     = Tau
          end if

          a_Hist(iLoop) = Tau
          if ( abs( ( DelEntr2 - DelEntr1 ) / DelEntr1 ) < ConvThreshold ) then
            exit search_entropy
          end if
        end do search_entropy
        if ( iLoop >= NLoop ) then
          write( 6, * ) 'ERROR: In ', trim( ModuleName )
          write( 6, * ) 'ERROR: In FindDelEntropyForTempEq'
          write( 6, * ) 'ERROR:   Loop does not converge'
          write( 6, * ) 'ERROR:   ', iLoop
          write( 6, * ) 'ERROR:   ', a_Hist
          stop
        end if

      end if

      Val = DelEntr

    end if


  end function IR_FindDelEntropyForTempEq

  !----------------------------------------------------------------------------
  ! Internal routine
  ! Find entropy difference for a specified equilibrium pressure
  !----------------------------------------------------------------------------

  function IR_FindDelEntropyForPressEq( PressEq ) &
    & result( Val )

    real(dp), intent(in) :: PressEq

    real(dp)             :: Val

    ! Local variables
    !
    real(dp) :: PiEq
    real(dp) :: DelEntr1
    real(dp) :: DelEntr2
    real(dp) :: DelEntr
    real(dp) :: Pi1
    real(dp) :: Pi2
    real(dp) :: Pi

    integer , parameter :: NLoop = 100
    integer             :: iLoop
    real(dp)            :: a_Hist( 0:NLoop )


    PiEq = PressEq / PressC

    if ( PiEq > 1.0d0 ) then

      write( 6, * ) 'ERROR: In ', trim( ModuleName )
      write( 6, * ) 'ERROR: Specified pressure is higher than the critical pressure.'
      write( 6, * ) '       ', PressEq, PressC
      stop

    else

      ! Initialization
      DelEntr1 = 1.0d-1
      DelEntr2 = 7.0d2

      DelEntr = DelEntr1
      Pi      = IR_vdWEOSRedPressAtPhaseEq( DelEntr/2.0d0 )
      Pi1     = Pi
      !
      DelEntr = DelEntr2
      Pi      = IR_vdWEOSRedPressAtPhaseEq( DelEntr/2.0d0 )
      Pi2     = Pi

      iLoop         = 0
      a_Hist(iLoop) = Pi1
      search_entropy : do iLoop = 1, NLoop
        DelEntr = ( DelEntr1 + DelEntr2 ) / 2.0d0
        Pi   = IR_vdWEOSRedPressAtPhaseEq( DelEntr/2.0d0 )

        if ( (Pi-PiEq) * (Pi1-PiEq) <= 0.0d0 ) then
          ! Target temperature is located between DelEntr1 and DelEntr.
          DelEntr2 = DelEntr
          Pi2      = Pi
        else
          ! Target temperature is located between DelEntr and DelEntr2.
          DelEntr1 = DelEntr
          Pi1      = Pi
        end if

        a_Hist(iLoop) = Pi
        if ( abs( ( DelEntr2 - DelEntr1 ) / DelEntr1 ) < ConvThreshold ) then
          exit search_entropy
        end if
      end do search_entropy
      if ( iLoop >= NLoop ) then
        write( 6, * ) 'ERROR: In ', trim( ModuleName )
        write( 6, * ) 'ERROR: In FindDelEntropyForPressEq'
        write( 6, * ) 'ERROR:   Loop does not converge'
        write( 6, * ) 'ERROR:   ', iLoop
        write( 6, * ) 'ERROR:   ', a_Hist
        stop
      end if

      Val = DelEntr

    end if


  end function IR_FindDelEntropyForPressEq

  !----------------------------------------------------------------------------
  ! Internal routine
  ! Calculate reduced dp/dT in phase equilibrium
  !----------------------------------------------------------------------------

  function IR_DRedPressDRedTempAtPhaseEq( X ) &
    & result( Val )

    real(dp), intent(in) :: X

    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: C
    real(dp) :: F
    real(dp) :: G
    real(dp) :: DCDX
    real(dp) :: DFDX
    real(dp) :: DGDX


    C    = IR_FuncC(X)
    F    = IR_FuncF(X)
    G    = IR_FuncG(X)
    DCDX = IR_FuncDCDX(X)
    DFDX = IR_FuncDFDX(X)
    DGDX = IR_FuncDGDX(X)
    Val = 8.0d0 * F &
      & * ( &
      &       ( 1.0d0 - 2.0d0 * F**2 ) * DFDX * G &
      &     - ( 1.0d0 - F**2 ) * F * DGDX         &
      &   ) &
      & / ( &
      &       ( 2.0d0 * F * DFDX + DFDX * C + F * DCDX ) * G &
      &     - 2.0d0 * F * ( F + C ) * DGDX                   &
      &   )


  end function IR_DRedPressDRedTempAtPhaseEq

  !----------------------------------------------------------------------------
  ! compressibility factor
  !----------------------------------------------------------------------------

  function ThermoPropCompFact( Dens, Temp ) &
    & result( CompFact )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: CompFact


    ! Local variables
    !
    real(DP) :: Delta
    real(DP) :: Tau
    real(DP) :: Pi


    Delta = Dens / DensC
    Tau   = Temp / TempC

    if ( Tau >= 1.0d0 ) then
      Pi = IR_vdWEOSRedPressOrg( Delta, Tau )
    else
      Pi = ThermoPropPress( Dens, Temp ) / PressC
    end if

    CompFact = 3.0d0 / 8.0d0 * Pi / ( Delta * Tau )


  end function ThermoPropCompFact

  !----------------------------------------------------------------------------
  ! pressure
  !----------------------------------------------------------------------------

  function ThermoPropPress( Dens, Temp ) &
    & result( Press )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Press


    ! Local variables
    !
    real(dp) :: Tau
    real(dp) :: Pi
    real(dp) :: Delta

    real(dp) :: DelEntr
    real(dp) :: PiEq
    real(dp) :: DeltaLiq
    real(dp) :: DeltaGas


    ! Initialization
    Delta = Dens / DensC
    Tau   = Temp / TempC


    if ( Tau >= 1.0d0 ) then

      Pi    = IR_vdWEOSRedPressOrg( Delta, Tau )

    else

      DelEntr = IR_FindDelEntropyForTempEq( Temp )
      PiEq    = IR_vdWEOSRedPressAtPhaseEq( DelEntr/2.0d0 )
      call IR_vdWEOSRedDensLimAtPhaseEq( DelEntr/2.0d0, DeltaLiq, DeltaGas )

      if ( ( DeltaGas <= Delta ) .and. ( Delta <= DeltaLiq ) ) then

        ! Region of density within phase equilibrium
        Pi = PiEq

      else

        ! Region of density lower or higher than phase equilibrium
        Pi = IR_vdWEOSRedPressOrg( Delta, Tau )

      end if

    end if

    Press = Pi * PressC


  end function ThermoPropPress

  !----------------------------------------------------------------------------
  ! density
  !----------------------------------------------------------------------------

  function ThermoPropDens( Press, Temp ) &
    & result( Dens )

    real(dp), intent(in) :: Press
    real(dp), intent(in) :: Temp
    real(dp)             :: Dens


    ! Local variables
    !
    real(DP) :: Tau
    real(DP) :: Pi
    real(DP) :: Delta

    real(dp) :: DeltaCur

    real(dp) :: FCur
    real(dp) :: DFDDelta

    real(dp) :: DeltaNext
    real(dp) :: DelDelta
    integer, parameter  :: NLoop = 100
    integer             :: iLoop
    real(dp)            :: a_Hist( 0:NLoop )

    real(dp) :: DelEntr
    real(dp) :: Delta1
    real(dp) :: Delta2
    real(dp) :: Pi1
    real(dp) :: Pi2
    real(dp) :: PiCur
    real(dp) :: DeltaGas
    real(dp) :: DeltaLiq
    real(dp) :: PiEq


    ! Initialization
    Tau   = Temp  / TempC
    Pi    = Press / PressC


    if ( Tau >= 1.0d0 ) then

      DeltaCur = 3.0d0 / 8.0d0 * Pi / Tau

      a_Hist(0) = DeltaCur
      search_density : do iLoop = 1, NLoop
        FCur =   3.0d0 * DeltaCur**3 &
          &    - 9.0d0 * DeltaCur**2 &
          &    + ( Pi + 8.0d0 * Tau ) * DeltaCur &
          &    - 3.0d0 * Pi
        DFDDelta =    9.0d0 * DeltaCur**2 &
          &        - 18.0d0 * DeltaCur    &
          &        + Pi + 8.0d0 * Tau
        DelDelta = ( 0.0d0 - FCur ) / DFDDelta

        DeltaNext = DeltaCur + DelDelta
        a_Hist( iLoop ) = DeltaNext
        if ( abs( ( DeltaNext - DeltaCur ) / DeltaCur ) < ConvThreshold ) then
          exit search_density
        end if
        DeltaCur = DeltaNext
      end do search_density
      if ( iLoop >= NLoop ) then
        write( 6, * ) 'ERROR: In ', trim( ModuleName )
        write( 6, * ) 'ERROR: In ThermoPropDens'
        write( 6, * ) 'ERROR:   Loop does not converge'
        write( 6, * ) 'ERROR:   ', Pi, Tau
        write( 6, * ) 'ERROR:   ', iLoop
        write( 6, * ) 'ERROR:   ', a_Hist
        stop
      end if

      Delta = DeltaNext

    else

      DelEntr = IR_FindDelEntropyForTempEq( Temp )
      PiEq    = IR_vdWEOSRedPressAtPhaseEq( DelEntr/2.0d0 )
      call IR_vdWEOSRedDensLimAtPhaseEq( DelEntr/2.0d0, DeltaLiq, DeltaGas )

      if ( Pi <= PiEq ) then
        ! Region of density lower than phase equilibrium
        Delta1   = DeltaGas
        DelDelta = 1.0d-1
      else
        ! Region of density higher than phase equilibrium
        Delta1   = DeltaLiq
        DelDelta = 1.0d1
      end if

      ! NOTE
      !   Pi1 calculated by IR_vdWEOSRedPressOrg as below should be the same
      !   as PiEq. It should be noticed that the relative accuracy is
      !   about 1.0e-3. 
      !   CHECK
!!$      Pi1 = IR_vdWEOSRedPressOrg( Delta1, Tau )
!!$      if ( abs( Pi1 - PiEq ) / Pi1 > 1.0d-3 ) then
!!$        write( 6, * ) 'ERROR: In ', trim( ModuleName )
!!$        write( 6, * ) 'ERROR: In ThermoPropDens'
!!$        write( 6, * ) '       Unexpectedly, PiEq is not equal to the pressure by EOS.'
!!$        write( 6, * ) '       ', Pi1, PiEq
!!$        stop
!!$      end if
      Pi1 = PiEq

      search_delta_1 : do
        Delta2 = Delta1 * DelDelta
        Delta2 = min( Delta2, 3.0d0-1.0d-10 )
        Pi2    = IR_vdWEOSRedPressOrg( Delta2, Tau )
        if ( ( Pi - Pi1 ) * ( Pi - Pi2 ) <= 0.0d0 ) then
          exit search_delta_1
        else
          Delta1 = Delta2
          Pi1    = Pi2
        end if
      end do search_delta_1

      ! bisection method
      a_Hist(0) = Delta1
      search_delta_2 : do iLoop = 1, NLoop
        DeltaCur = ( Delta1 + Delta2 ) / 2.0d0
        PiCur    = IR_vdWEOSRedPressOrg( DeltaCur, Tau )
        if ( ( Pi - PiCur ) * ( Pi - Pi1 ) <= 0.0d0 ) then
          Delta2 = DeltaCur
          Pi2    = PiCur
        else
          Delta1 = DeltaCur
          Pi1    = PiCur
        end if
        a_Hist(iLoop) = DeltaCur
        if ( abs( ( Delta2 - Delta1 ) / Delta1 ) < ConvThreshold ) then
          exit search_delta_2
        end if
      end do search_delta_2
      if ( iLoop >= NLoop ) then
        if ( Pi <= PiEq ) then
          write( 6, * ) 'ERROR: In ', trim( ModuleName )
          write( 6, * ) 'ERROR: In ThermoPropDens, lower density region'
        else
          write( 6, * ) 'ERROR: In ', trim( ModuleName )
          write( 6, * ) 'ERROR: In ThermoPropDens, higher density region'
        end if
        write( 6, * ) 'ERROR:   Loop does not converge'
        write( 6, * ) 'ERROR:   ', Pi, Tau
        write( 6, * ) 'ERROR:   ', iLoop
        write( 6, * ) 'ERROR:   ', a_Hist
        stop
      end if

      Delta = DeltaCur

    end if

    Dens = Delta * DensC

  end function ThermoPropDens

  !----------------------------------------------------------------------------
  ! temperature
  !----------------------------------------------------------------------------

  function ThermoPropTemp( Press, Dens ) &
    & result( Temp )

    real(dp), intent(in) :: Press
    real(dp), intent(in) :: Dens
    real(dp)             :: Temp


    ! Local variables
    !
    real(DP) :: Delta
    real(DP) :: Tau
    real(DP) :: Pi
    real(DP) :: DelEntr
    real(DP) :: DeltaGas
    real(DP) :: DeltaLiq


    Delta = Dens  / DensC
    Pi    = Press / PressC


    if ( Pi >= 1.0d0 ) then

      Tau = IR_vdWEOSRedTempOrg( Delta, Pi )

    else

      DelEntr = IR_FindDelEntropyForPressEq( Press )
      call IR_vdWEOSRedDensLimAtPhaseEq( DelEntr/2.0d0, DeltaLiq, DeltaGas )

      if ( ( DeltaGas <= Delta ) .and. ( Delta <= DeltaLiq ) ) then

        ! Region of density in phase equilibrium
        Tau = IR_vdWEOSRedTempAtPhaseEq( DelEntr/2.0d0 )

      else

        ! Region of density lower or higher than phase equilibrium
        Tau = IR_vdWEOSRedTempOrg( Delta, Pi )

      end if

    end if

    Temp = Tau * TempC


  end function ThermoPropTemp

  !----------------------------------------------------------------------------
  ! internal energy
  !----------------------------------------------------------------------------

  function ThermoPropInternalEnergy( Dens, Temp ) &
    & result( Val )

    use thermo_prop_mix_gs2016_co2n2, only : &
      & ThermoPropInternalEnergyGS2016 => ThermoPropInternalEnergy


    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: Delta
    real(dp) :: Tau
    real(dp) :: IGIntEne
    real(dp) :: DelEntr
    real(dp) :: DeltaLiq
    real(dp) :: DeltaGas
    real(dp) :: PiEq
    real(dp) :: DPDTEq


    Delta = Dens / DensC
    Tau   = Temp / TempC

    IGIntEne = ThermoPropInternalEnergyGS2016( Dens, Temp )

    if ( Tau >= 1.0d0 ) then

      Val = IGIntEne - 3.0d0 * Delta * PressC / DensC

    else

      DelEntr = IR_FindDelEntropyForTempEq( Temp )
      call IR_vdWEOSRedDensLimAtPhaseEq( DelEntr/2.0d0, DeltaLiq, DeltaGas )

      if ( Delta <= DeltaGas ) then

        ! Region of density lower than that in phase equilibrium

        Val = IGIntEne &
          & - 3.0d0 * Delta * PressC / DensC

      else if ( Delta <= DeltaLiq ) then

        ! Region of density in phase equilibrium

        PiEq = IR_vdWEOSRedPressAtPhaseEq( DelEntr/2.0d0 )
        DPDTEq = IR_DRedPressDRedTempAtPhaseEq( DelEntr/2.0d0 )

        Val = IGIntEne &
          & - 3.0d0 * DeltaGas * PressC / DensC &
          & - ( PiEq - Tau * DPDTEq ) &
          &   * ( 1.0d0 / Delta - 1.0d0 / DeltaGas ) &
          &   * PressC / DensC


!!$        Val = -1.0d0
!!$
!!$        write( 6, * ) 'ERROR: In ', trim( ModuleName )
!!$        write( 6, * ) 'ERROR: In ThermoPropInternalEnergy'
!!$        write( 6, * ) '       This range is not supported.'
!!$        write( 6, * ) '       Range of Delta : ', DeltaGas, Delta, DeltaGas
!!$        stop

      else

        ! Region of density higher than that in phase equilibrium

        PiEq = IR_vdWEOSRedPressAtPhaseEq( DelEntr/2.0d0 )
        DPDTEq = IR_DRedPressDRedTempAtPhaseEq( DelEntr/2.0d0 )

        Val = IGIntEne &
          & - 3.0d0 * DeltaGas * PressC / DensC &
          & - ( PiEq - Tau * DPDTEq ) &
          &   * ( 1.0d0 / DeltaLiq - 1.0d0 / DeltaGas ) &
          &   * PressC / DensC
        Val = Val &
          & - 3.0d0 * ( Delta - DeltaLiq ) * PressC / DensC


!!$        Val = -1.0d0
!!$
!!$        write( 6, * ) 'ERROR: In ', trim( ModuleName )
!!$        write( 6, * ) 'ERROR: In ThermoPropInternalEnergy'
!!$        write( 6, * ) '       This range is not supported.'
!!$        write( 6, * ) '       Range of Delta : ', DeltaGas, Delta, DeltaGas
!!$        stop

      end if


    end if

    !--------------------------------------------------------------------------

  end function ThermoPropInternalEnergy

  !----------------------------------------------------------------------------
  ! enthalpy
  !----------------------------------------------------------------------------

  function ThermoPropEnthalpy( Dens, Temp ) &
    & result( Val )


    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: IntEne
    real(dp) :: Press


    IntEne = ThermoPropInternalEnergy( Dens, Temp )
    Press  = ThermoPropPress( Dens, Temp )

    Val = IntEne + Press / Dens


  end function ThermoPropEnthalpy

  !----------------------------------------------------------------------------
  ! isobaric heat capacity (specific heat on constant pressure)
  !----------------------------------------------------------------------------

  function ThermoPropCp( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !


    Val = ThermoPropCv( Dens, Temp )
    Val = Val &
      & + Temp / Dens &
      &   * ThermoPropDPressDTempConstDens( Dens, Temp ) &
      &   * ThermoPropThermoExpCoef( Dens, Temp )


  end function ThermoPropCp

  !----------------------------------------------------------------------------
  ! isochoric heat capacity (specific heat on constant volume)
  !----------------------------------------------------------------------------

  function ThermoPropCv( Dens, Temp ) &
    & result( Val )

    use thermo_prop_mix_gs2016_co2n2, only : &
      & ThermoPropCvGS2016 => ThermoPropCv

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !


    ! Cv of ideal gas, and non-ideal gas component is zero.
    Val = ThermoPropCvGS2016( Dens, Temp )


  end function ThermoPropCv

  !----------------------------------------------------------------------------
  ! DPDRho
  !----------------------------------------------------------------------------

  function ThermoPropDPressDDensConstTemp( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: Delta
    real(dp) :: Tau
    real(dp) :: DelEntr
    real(dp) :: DeltaGas
    real(dp) :: DeltaLiq


    Delta = Dens / DensC
    Tau   = Temp / TempC


    if ( Tau >= 1.0d0 ) then

      Val = DRedPressDRedDensConstTemp( Delta, Tau )

    else

      DelEntr = IR_FindDelEntropyForTempEq( Temp )
      call IR_vdWEOSRedDensLimAtPhaseEq( DelEntr/2.0d0, DeltaLiq, DeltaGas )

      if ( ( DeltaGas <= Delta ) .and. ( Delta <= DeltaLiq ) ) then

        ! Region of density in phase equilibrium
        Val = 0.0d0

        ! The value would be zero.
        ! But, now, this program stops and the value is not returned.

        write( 6, * ) 'ERROR: In ', trim( ModuleName )
        write( 6, * ) 'ERROR: In ThermoPropDPressDDensConstTemp'
        write( 6, * ) '       This range is not supported.'
        stop

      else

        ! Region of density lower or higher than phase equilibrium
        Val = DRedPressDRedDensConstTemp( Delta, Tau )

      end if

    end if

    Val = Val * PressC / DensC


    !--------------------------------------------------------------------------

  contains

    !--------------------------------------------------------------------------

    function DRedPressDRedDensConstTemp( Delta, Tau ) &
      & result( Val )

      real(dp), intent(in) :: Delta
      real(dp), intent(in) :: Tau
      real(dp)             :: Val


      Val = 24.0d0 * Tau / ( 3.0d0 - Delta )**2 - 6.0d0 * Delta


    end function DRedPressDRedDensConstTemp

    !--------------------------------------------------------------------------

  end function ThermoPropDPressDDensConstTemp

  !----------------------------------------------------------------------------
  ! DPDT
  !----------------------------------------------------------------------------

  function ThermoPropDPressDTempConstDens( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: Delta
    real(dp) :: Tau
    real(dp) :: DelEntr
    real(dp) :: DeltaLiq
    real(dp) :: DeltaGas


    Delta = Dens / DensC
    Tau   = Temp / TempC

    if ( Tau >= 1.0d0 ) then

      Val = DRedPressDRedTempConstDens( Delta )

    else

      DelEntr = IR_FindDelEntropyForTempEq( Temp )
      call IR_vdWEOSRedDensLimAtPhaseEq( DelEntr/2.0d0, DeltaLiq, DeltaGas )

      if ( ( DeltaGas <= Delta ) .and. ( Delta <= DeltaLiq ) ) then

        ! Region of density in phase equilibrium

        Val = IR_DRedPressDRedTempAtPhaseEq( DelEntr/2.0d0 )

!!$        Val = -1.0d0
!!$        write( 6, * ) 'ERROR: In ', trim( ModuleName )
!!$        write( 6, * ) 'ERROR: In ThermoPropDPressDTempConstDens'
!!$        write( 6, * ) '       This range is not supported.'
!!$        stop

      else

        ! Region of density lower or higher than phase equilibrium

        Val = DRedPressDRedTempConstDens( Delta )

      end if


    end if

    Val = Val * PressC / TempC


    !--------------------------------------------------------------------------

  contains

    !--------------------------------------------------------------------------

    function DRedPressDRedTempConstDens( Delta ) &
      & result( Val )

      real(dp), intent(in) :: Delta

      real(dp)             :: Val


      Val = 8.0d0 * Delta / ( 3.0d0 - Delta )


    end function DRedPressDRedTempConstDens

    !--------------------------------------------------------------------------

  end function ThermoPropDPressDTempConstDens

  !----------------------------------------------------------------------------
  ! DHDRho
  !----------------------------------------------------------------------------

  function ThermoPropDEnthalpyDDensConstTemp( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: Delta
    real(dp) :: Tau
    real(dp) :: DelEntr
    real(dp) :: DeltaLiq
    real(dp) :: DeltaGas


    Delta = Dens / DensC
    Tau   = Temp / TempC

    if ( Tau >= 1.0d0 ) then

      Val = DEnthalpyDDensConstTemp( Delta, Tau )

    else

      DelEntr = IR_FindDelEntropyForTempEq( Temp )
      call IR_vdWEOSRedDensLimAtPhaseEq( DelEntr/2.0d0, DeltaLiq, DeltaGas )

      if ( Delta <= DeltaGas ) then

        ! Region of density lower than that in phase equilibrium

        Val = DEnthalpyDDensConstTemp( Delta, Tau )

      else if ( Delta <= DeltaLiq ) then

        ! Region of density in phase equilibrium

        Val = - IR_DRedPressDRedTempAtPhaseEq( DelEntr/2.0d0 )
        Val = Val * Tau / ( Delta**2 )
        Val = Val * PressC / DensC**2

!!$        Val = -1.0d0
!!$
!!$        write( 6, * ) 'ERROR: In ', trim( ModuleName )
!!$        write( 6, * ) 'ERROR: In ThermoPropDEnthalpyDDensConstTemp'
!!$        write( 6, * ) '       This range is not supported.'
!!$        stop

      else

        ! Region of density higher than that in phase equilibrium

        Val = DEnthalpyDDensConstTemp( Delta, Tau )

!!$        Val = -1.0d0
!!$
!!$        write( 6, * ) 'ERROR: In ', trim( ModuleName )
!!$        write( 6, * ) 'ERROR: In ThermoPropDEnthalpyDDensConstTemp'
!!$        write( 6, * ) '       This range is not supported.'
!!$        stop

      end if

    end if

    !--------------------------------------------------------------------------

  contains

    !--------------------------------------------------------------------------

    function DEnthalpyDDensConstTemp( Delta, Tau ) &
      & result( Val )

      real(dp), intent(in) :: Delta
      real(dp), intent(in) :: Tau
      real(dp)             :: Val


      Val = 8.0d0 * Tau / ( 3.0d0 - Delta )**2 - 6.0d0
      Val = Val * PressC / DensC**2


    end function DEnthalpyDDensConstTemp

    !--------------------------------------------------------------------------

  end function ThermoPropDEnthalpyDDensConstTemp

  !----------------------------------------------------------------------------
  ! DHDT
  !----------------------------------------------------------------------------

  function ThermoPropDEnthalpyDTempConstDens( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: Delta
    real(dp) :: Tau
    real(dp) :: DelEntr
    real(dp) :: DeltaLiq
    real(dp) :: DeltaGas


    Delta = Dens / DensC
    Tau   = Temp / TempC

    if ( Tau >= 1.0d0 ) then

      Val = DEnthalpyDTempConstDens( Delta, Tau )

    else

      DelEntr = IR_FindDelEntropyForTempEq( Temp )
      call IR_vdWEOSRedDensLimAtPhaseEq( DelEntr/2.0d0, DeltaLiq, DeltaGas )

      if ( Delta <= DeltaGas ) then

        ! Region of density lower than that in phase equilibrium

        Val = DEnthalpyDTempConstDens( Delta, Tau )

      else if ( Delta <= DeltaLiq ) then

        ! Region of density in phase equilibrium

        Val = -1.0d0

        write( 6, * ) 'ERROR: In ', trim( ModuleName )
        write( 6, * ) 'ERROR: In ThermoPropDEnthalpyDTempConstDens'
        write( 6, * ) '       This range is not supported.'
        stop

      else

        ! Region of density higher than that in phase equilibrium

        ! This would be right.
        Val = DEnthalpyDTempConstDens( Delta, Tau )

!!$        Val = -1.0d0
!!$
!!$        write( 6, * ) 'ERROR: In ', trim( ModuleName )
!!$        write( 6, * ) 'ERROR: In ThermoPropDEnthalpyDTempConstDens'
!!$        write( 6, * ) '       This range is not supported.'
!!$        stop

      end if

    end if

    !--------------------------------------------------------------------------

  contains

    !--------------------------------------------------------------------------

    function DEnthalpyDTempConstDens( Delta, Tau ) &
      & result( Val )

      real(dp), intent(in) :: Delta
      real(dp), intent(in) :: Tau
      real(dp)             :: Val


      Val = ThermoPropCv( Delta*DensC, Tau*TempC )
      Val = Val + 8.0d0 / ( 3.0d0 - Delta ) * PressC / ( DensC * TempC )


    end function DEnthalpyDTempConstDens

    !--------------------------------------------------------------------------


  end function ThermoPropDEnthalpyDTempConstDens

  !----------------------------------------------------------------------------
  ! thermal expansion coefficient
  !----------------------------------------------------------------------------

  function ThermoPropThermoExpCoef( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !

    Val =   1.0d0 / Dens &
      &   * ThermoPropDPressDTempConstDens( Dens, Temp ) &
      &   / ThermoPropDPressDDensConstTemp( Dens, Temp )


  end function ThermoPropThermoExpCoef

  !----------------------------------------------------------------------------
  ! adiabatic lapse rate (DT/DP)
  !----------------------------------------------------------------------------

  function ThermoPropAdLapRateDTDP( Dens, Temp ) &
    & result( Val )

    real(dp), intent(in) :: Dens
    real(dp), intent(in) :: Temp
    real(dp)             :: Val


    ! Local variables
    !
    real(dp) :: Cp
    real(dp) :: ExpCoef


    Cp       = ThermoPropCp( Dens, Temp )
    ExpCoef  = ThermoPropThermoExpCoef( Dens, Temp )

    Val = Temp / ( Cp * Dens ) * ExpCoef


  end function ThermoPropAdLapRateDTDP

  !----------------------------------------------------------------------------
  ! temperature and density from pressure and enthalpy
  !----------------------------------------------------------------------------

  subroutine ThermoPropDensTempFromPressEnthalpy( &
    & Press, Enthalpy, & ! (in)
    & Dens, Temp       & ! (out)
    & )

    real(dp), intent(in ) :: Press
    real(dp), intent(in ) :: Enthalpy
    real(dp), intent(out) :: Dens
    real(dp), intent(out) :: Temp


    ! Local variables
    !
    real(dp) :: Temp1
    real(dp) :: Temp2
!!$    real(dp) :: Dens1
!!$    real(dp) :: Dens2
    real(dp) :: Enthalpy1
    real(dp) :: Enthalpy2

    real(dp) :: DensCur
    real(dp) :: DensNext
    real(dp) :: TempCur
    real(dp) :: TempNext
    real(dp) :: PressCur
    real(dp) :: EnthalpyCur
    real(dp) :: DPDRhoCT
    real(dp) :: DPDTCRho
    real(dp) :: DHDRhoCT
    real(dp) :: DHDTCRho
    real(dp) :: DelDens
    real(dp) :: DelTemp
    integer, parameter  :: NLoop = 100
    integer             :: iLoop
    real(dp)            :: a_HistDens( 0:NLoop )
    real(dp)            :: a_HistTemp( 0:NLoop )


    ! Initialization
    !   find temperature with a given pressure

    ! Bisectional method
    Temp1     = 50.0d0
    DensCur   = ThermoPropDens( Press, Temp1 )
    Enthalpy1 = ThermoPropEnthalpy( DensCur, Temp1 )
    if ( Enthalpy < Enthalpy1 ) then
      write( 6, * ) 'ERROR: In ', trim( ModuleName )
      write( 6, * ) 'ERROR: In ThermoPropDensTempFromPressEnthalpy'
      write( 6, * ) 'ERROR:   Temperature seems to be too low.'
      stop
    end if
    search_init_temperature : do
      if ( Temp1 < 100.0d0 ) then
        Temp2     = Temp1 + 10.0d0
      else
        Temp2     = Temp1 + 100.0d0
      end if
      DensCur   = ThermoPropDens( Press, Temp2 )
      Enthalpy2 = ThermoPropEnthalpy( DensCur, Temp2 )
      if ( Enthalpy < Enthalpy2 ) then
        exit
      end if
      Temp1 = Temp2
    end do search_init_temperature
    bisectional_method : do
      TempCur     = ( Temp1 + Temp2 ) / 2.0d0
      DensCur     = ThermoPropDens( Press, TempCur )
      EnthalpyCur = ThermoPropEnthalpy( DensCur, TempCur )
      if ( Enthalpy < EnthalpyCur ) then
        Temp2 = TempCur
      else
        Temp1 = TempCur
      end if
      if ( abs( ( Temp2 - Temp1 ) / Temp1 ) < ConvThreshold ) then
        exit bisectional_method
      end if
    end do bisectional_method
    TempCur = ( Temp1 + Temp2 ) / 2.0d0


    ! Newton's method
    ! NOTE: This does not work well in some cases, i.e., case of liquid.
!!$    TempCur = 300.0d0
!!$    a_HistTemp(0) = TempCur
!!$    search_temperature : do iLoop = 1, NLoop
!!$      Temp1     = TempCur
!!$      Temp2     = Temp1 + 1.0d-1
!!$      Dens1     = ThermoPropDens( Press, Temp1 )
!!$      Dens2     = ThermoPropDens( Press, Temp2 )
!!$      Enthalpy1 = ThermoPropEnthalpy( Dens1, Temp1 )
!!$      Enthalpy2 = ThermoPropEnthalpy( Dens2, Temp2 )
!!$      ! (DHDT)_p
!!$      DHDT = ( Enthalpy2 - Enthalpy1 ) / ( Temp2 - Temp1 )
!!$      !
!!$      EnthalpyCur = Enthalpy1
!!$      DelTemp = ( Enthalpy - EnthalpyCur ) / DHDT
!!$      TempNext = TempCur + DelTemp
!!$      TempNext = max( TempNext, 50.0d0 )
!!$      a_HistTemp( iLoop ) = TempNext
!!$      write( 6, * ) iLoop, TempCur, TempNext
!!$      if ( abs( ( TempNext - TempCur ) / TempCur ) < ConvThreshold ) then
!!$        exit search_temperature
!!$      end if
!!$      TempCur = TempNext
!!$    end do search_temperature
!!$    if ( iLoop >= NLoop ) then
!!$      write( 6, * ) 'ERROR: In ', trim( ModuleName )
!!$      write( 6, * ) 'ERROR: In ThermoPropDensTempFromPressEnthalpy'
!!$      write( 6, * ) 'ERROR:   Loop for temperature does not converge'
!!$      write( 6, * ) 'ERROR:   ', iLoop
!!$      write( 6, * ) 'ERROR:   ', a_HistTemp
!!$      stop
!!$    end if
!!$    TempCur = TempNext


    DensCur = ThermoPropDens( Press, TempCur )

    a_HistDens(0) = DensCur
    a_HistTemp(0) = TempCur
    search_density_temperature : do iLoop = 1, NLoop
      PressCur    =                                       &
        & ThermoPropCompFact( DensCur, TempCur ) &
        & * DensCur * GasConst * TempCur
      EnthalpyCur = ThermoPropEnthalpy( DensCur, TempCur )

      DPDRhoCT = ThermoPropDPressDDensConstTemp( DensCur, TempCur )
      DPDTCRho = ThermoPropDPressDTempConstDens( DensCur, TempCur )

      DHDRhoCT = ThermoPropDEnthalpyDDensConstTemp( DensCur, TempCur )
      DHDTCRho = ThermoPropDEnthalpyDTempConstDens( DensCur, TempCur )
      DelTemp =                                         &
        &   (   ( Enthalpy - EnthalpyCur ) * DPDRhoCT   &
        &     - ( Press    - PressCur    ) * DHDRhoCT ) &
        & / ( DHDTCRho * DPDRhoCT - DHDRhoCT * DPDTCRho )
      DelDens = &
        & ( Enthalpy - EnthalpyCur - DHDTCRho * DelTemp ) / DHDRhoCT
      DensNext = DensCur + DelDens
      TempNext = TempCur + DelTemp

      a_HistDens( iLoop ) = DensNext
      a_HistTemp( iLoop ) = TempNext
!!$      write( 6, * ) iLoop, DensCur, DensNext, TempCur, TempNext
      if ( ( abs( ( DensNext - DensCur ) / DensCur ) < ConvThreshold ) .and. &
        &  ( abs( ( TempNext - TempCur ) / TempCur ) < ConvThreshold ) ) then
        exit search_density_temperature
      end if
      DensCur  = DensNext
      TempCur  = TempNext
    end do search_density_temperature
    if ( iLoop >= NLoop ) then
      write( 6, * ) 'ERROR: In ', trim( ModuleName )
      write( 6, * ) 'ERROR: In ThermoPropDensTempFromPressEnthalpy'
      write( 6, * ) 'ERROR:   Loop does not converge'
      write( 6, * ) 'ERROR:   ', Press, Enthalpy
      write( 6, * ) 'ERROR:   ', iLoop
      write( 6, * ) 'ERROR:   ', a_HistDens
      write( 6, * ) 'ERROR:   ', a_HistTemp
      stop
    end if

    Dens = DensNext
    Temp = TempNext


  end subroutine ThermoPropDensTempFromPressEnthalpy

  !----------------------------------------------------------------------------

end module thermo_prop_vdWEOS
